"""
Common interface for multiple e-mail methods

``outgoing`` provides a common interface to multiple different e-mail sending
methods (SMTP, sendmail, mbox, etc.).  Just construct a sender from a
configuration file or object, pass it an ``EmailMessage`` instance, and let the
magical internet daemons take care of the rest.

``outgoing`` itself provides support for only basic sending methods; additional
methods are provided by other packages.

Visit <https://github.com/jwodder/outgoing> or <https://outgoing.rtfd.io> for
more information.
"""

__version__ = "0.2.4"
__author__ = "John Thorvald Wodder II"
__author_email__ = "outgoing@varonathe.org"
__license__ = "MIT"
__url__ = "https://github.com/jwodder/outgoing"

from .config import (
    DirectoryPath,
    FilePath,
    NetrcConfig,
    Password,
    Path,
    StandardPassword,
)
from .core import (
    DEFAULT_CONFIG_SECTION,
    Sender,
    from_config_file,
    from_dict,
    get_default_configpath,
    lookup_netrc,
    resolve_password,
)
from .errors import (
    Error,
    InvalidConfigError,
    InvalidPasswordError,
    MissingConfigError,
    NetrcLookupError,
)
from .util import OpenClosable, resolve_path

__all__ = [
    "DEFAULT_CONFIG_SECTION",
    "DirectoryPath",
    "Error",
    "FilePath",
    "InvalidConfigError",
    "InvalidPasswordError",
    "MissingConfigError",
    "NetrcConfig",
    "NetrcLookupError",
    "OpenClosable",
    "Password",
    "Path",
    "Sender",
    "StandardPassword",
    "from_config_file",
    "from_dict",
    "get_default_configpath",
    "lookup_netrc",
    "resolve_password",
    "resolve_path",
]



from email.message import EmailMessage
import logging
from pathlib import Path
import pytest
from outgoing import Sender, from_dict
from outgoing.senders.null import NullSender


def test_null_construct(tmp_path: Path) -> None:
    sender = from_dict({"method": "null"}, configpath=tmp_path / "foo.toml")
    assert isinstance(sender, Sender)
    assert isinstance(sender, NullSender)
    assert sender.dict() == {"configpath": tmp_path / "foo.toml"}


def test_null_send(caplog: pytest.LogCaptureFixture, test_email1: EmailMessage) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    sender = from_dict({"method": "null"})
    with sender as s:
        assert sender is s
        sender.send(test_email1)
    assert caplog.record_tuples == [
        (
            "outgoing.senders.null",
            logging.INFO,
            f"Discarding e-mail {test_email1['Subject']!r}",
        )
    ]


def test_null_send_no_context(test_email1: EmailMessage) -> None:
    sender = from_dict({"method": "null"})
    sender.send(test_email1)



from collections.abc import Mapping as MappingABC
from email.message import EmailMessage
import inspect
import json
from netrc import netrc
import os
from pathlib import Path
import sys
from types import TracebackType
from typing import Any, Mapping, Optional, Tuple, Type, TypeVar, Union, cast
import appdirs
import entrypoints
import tomli
from . import errors
from .util import AnyPath

if sys.version_info[:2] >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable

DEFAULT_CONFIG_SECTION = "outgoing"

SENDER_GROUP = "outgoing.senders"

PASSWORD_SCHEME_GROUP = "outgoing.password_schemes"

S = TypeVar("S", bound="Sender")


@runtime_checkable
class Sender(Protocol):
    """
    `Sender` is a `~typing.Protocol` implemented by sender objects.  The
    protocol requires the following behavior:

    - Sender objects can be used as context managers, and their ``__enter__``
      methods return ``self``.

    - Within its own context, calling a sender's ``send(msg:
      email.message.EmailMessage)`` method sends the given e-mail.
    """

    def __enter__(self: S) -> S:
        ...

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        ...

    def send(self, msg: EmailMessage) -> Any:
        """ Send ``msg`` or raise an exception if that's not possible """
        ...


def get_default_configpath() -> Path:
    """
    Returns the location of the default config file (regardless of whether it
    exists) as a `pathlib.Path` object
    """
    return Path(appdirs.user_config_dir("outgoing", "jwodder"), "outgoing.toml")


def from_config_file(
    path: Optional[AnyPath] = None,
    section: Optional[str] = DEFAULT_CONFIG_SECTION,
    fallback: bool = True,
) -> Sender:
    """
    Read configuration from the table/field ``section`` (default
    "``outgoing``") in the file at ``path`` (default: the path returned by
    `get_default_configpath()`) and construct a sender object from the
    specification.  The file may be either TOML or JSON (type detected based on
    file extension).  If ``section`` is `None`, the entire file, rather than
    only a single field, is used as the configuration.  If ``fallback`` is
    true, the file is not the default config file, and the file either does not
    exist or does not contain the given section, fall back to reading from the
    default section of the default config file.

    :raises InvalidConfigError: if the configuration is invalid
    :raises MissingConfigError: if no configuration file or section is present
    """
    if path is None:
        configpath = get_default_configpath()
    else:
        configpath = Path(os.fsdecode(path))
    data: Any
    try:
        if configpath.suffix == ".toml":
            with configpath.open("rb") as fb:
                data = tomli.load(fb)
        elif configpath.suffix == ".json":
            with configpath.open("r") as fp:
                data = json.load(fp)
        else:
            raise errors.InvalidConfigError(
                "Unsupported file extension",
                configpath=configpath,
            )
    except FileNotFoundError:
        data = None
    if data is not None and section is not None:
        if not isinstance(data, MappingABC):
            raise errors.InvalidConfigError(
                "Top-level structure must be a dict/object",
                configpath=configpath,
            )
        data = data.get(section)
    if data is None:
        if fallback and configpath != get_default_configpath():
            try:
                return from_config_file(fallback=False)
            except errors.MissingConfigError as e:
                e.configpaths.append(configpath)
                raise e
        else:
            raise errors.MissingConfigError([configpath])
    if not isinstance(data, MappingABC):
        raise errors.InvalidConfigError(
            "Section must be a dict/object",
            configpath=configpath,
        )
    return from_dict(data, configpath=configpath)


def from_dict(
    data: Mapping[str, Any],
    configpath: Optional[AnyPath] = None,
) -> Sender:
    """
    Construct a sender object using the given ``data`` as the configuration.
    If ``configpath`` is given, any paths in the ``data`` will be resolved
    relative to ``configpath``'s parent directory; otherwise, they will be
    resolved relative to the current directory.

    ``data`` should not contain a ``"configpath"`` key; such an entry will be
    discarded.

    :raises InvalidConfigError: if the configuration is invalid
    """
    try:
        method = data["method"]
    except KeyError:
        raise errors.InvalidConfigError(
            "Required 'method' field not present",
            configpath=configpath,
        )
    if "configpath" in data:
        # TODO: Emit warning
        data = dict(data)
        data.pop("configpath", None)
    try:
        ep = entrypoints.get_single(SENDER_GROUP, method)
    except entrypoints.NoSuchEntryPoint:
        raise errors.InvalidConfigError(
            f"Unsupported method {method!r}",
            configpath=configpath,
        )
    sender_cls = ep.load()
    try:
        return cast(Sender, sender_cls(configpath=configpath, **data))
    except (TypeError, ValueError) as e:
        raise errors.InvalidConfigError(str(e), configpath=configpath)
    except errors.InvalidConfigError as e:
        if e.configpath is None:
            e.configpath = configpath
        raise e


def resolve_password(
    password: Union[str, Mapping[str, Any]],
    host: Optional[str] = None,
    username: Optional[str] = None,
    configpath: Union[str, Path, None] = None,
) -> str:
    """
    Resolve a configuration password value.  If ``password`` is a string, it is
    returned unchanged.  Otherwise, it must be a mapping with exactly one
    element; the key is used as the name of the password scheme, and the value
    is passed to the corresponding function for retrieving the password.

    When resolving a password field in an ``outgoing`` configuration structure,
    the configpath and any host/service or username values from the
    configuration (or host/service/username constants specific to the sending
    method) should be passed into this function so that they can be made
    available to any password scheme functions that need them.

    :raises InvalidPasswordError:
        if ``password`` is invalid or cannot be resolved
    """
    if isinstance(password, str):
        return password
    elif len(password) != 1:
        raise errors.InvalidPasswordError(
            "Password must be either a string or an object with exactly one field"
        )
    ((scheme, spec),) = password.items()
    try:
        ep = entrypoints.get_single(PASSWORD_SCHEME_GROUP, scheme)
    except entrypoints.NoSuchEntryPoint:
        raise errors.InvalidPasswordError(
            f"Unsupported password scheme {scheme!r}",
            configpath=configpath,
        )
    scheme_func = ep.load()
    available_kwargs = {
        "host": host,
        "username": username,
        "configpath": configpath,
    }
    kwargs = {}
    sig = inspect.signature(scheme_func)
    for param in sig.parameters.values():
        if (
            param.kind in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY)
            and param.name in available_kwargs
        ):
            kwargs[param.name] = available_kwargs[param.name]
        elif param.kind is param.VAR_KEYWORD:
            kwargs.update(available_kwargs)
    try:
        return cast(str, scheme_func(spec=spec, **kwargs))
    except (TypeError, ValueError) as e:
        raise errors.InvalidPasswordError(str(e), configpath=configpath)
    except errors.InvalidPasswordError as e:
        if e.configpath is None:
            e.configpath = configpath
        raise e


def lookup_netrc(
    host: str, username: Optional[str] = None, path: Optional[AnyPath] = None
) -> Tuple[str, str]:
    """
    Look up the entry for ``host`` in the netrc file at ``path`` (default:
    :file:`~/.netrc`) and return a pair of the username & password.  If
    ``username`` is specified and it does not equal the username in the file, a
    `NetrcLookupError` is raised.

    :raises NetrcLookupError:
        if no entry for ``host`` or the default entry is present in the netrc
        file; or if ``username`` differs from the username in the netrc file
    :raises netrc.NetrcParseError: if the `netrc` module encounters an error
    """
    if path is None:
        rc = netrc()
    else:
        rc = netrc(os.fsdecode(path))
    auth = rc.authenticators(host)
    if auth is None:
        raise errors.NetrcLookupError(
            f"No entry for {host!r} or default found in netrc file"
        )
    elif username is not None and auth[0] != username:
        raise errors.NetrcLookupError(
            f"Username mismatch in netrc: expected {username!r},"
            f" but netrc says {auth[0]!r}"
        )
    password = auth[2]
    if password is None:  # pragma: no cover
        # mypy says this can happen, but the actual implementation in CPython
        # says otherwise.
        raise errors.NetrcLookupError("No password given in netrc entry")
    return (auth[0], password)



from email.message import EmailMessage
import logging
from mailbox import Babyl
from pathlib import Path
from mailbits import email2dict
import pytest
from outgoing import Sender, from_dict
from outgoing.senders.mailboxes import BabylSender


def test_babyl_construct(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "babyl",
            "path": "inbox",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    assert isinstance(sender, Sender)
    assert isinstance(sender, BabylSender)
    assert sender.dict() == {
        "configpath": tmp_path / "foo.txt",
        "path": tmp_path / "inbox",
    }
    assert sender._mbox is None


def test_babyl_send_new_path(
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    tmp_path: Path,
) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "babyl",
            "path": "inbox",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender as s:
        assert sender is s
        sender.send(test_email1)
    inbox = Babyl("inbox")
    inbox.lock()
    msgs = list(inbox)
    inbox.close()
    assert len(msgs) == 1
    assert email2dict(test_email1) == email2dict(msgs[0])
    assert caplog.record_tuples == [
        (
            "outgoing.senders.mailboxes",
            logging.DEBUG,
            f"Opening Babyl mailbox at {tmp_path/'inbox'}",
        ),
        (
            "outgoing.senders.mailboxes",
            logging.INFO,
            f"Adding e-mail {test_email1['Subject']!r} to Babyl mailbox at"
            f" {tmp_path/'inbox'}",
        ),
        (
            "outgoing.senders.mailboxes",
            logging.DEBUG,
            f"Closing Babyl mailbox at {tmp_path/'inbox'}",
        ),
    ]


def test_babyl_send_extant_path(
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    test_email2: EmailMessage,
    tmp_path: Path,
) -> None:
    monkeypatch.chdir(tmp_path)
    inbox = Babyl("inbox")
    inbox.lock()
    inbox.add(test_email1)
    inbox.close()
    sender = from_dict(
        {
            "method": "babyl",
            "path": "inbox",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender:
        sender.send(test_email2)
    inbox = Babyl("inbox")
    inbox.lock()
    msgs = list(inbox)
    inbox.close()
    assert len(msgs) == 2
    assert email2dict(test_email1) == email2dict(msgs[0])
    assert email2dict(test_email2) == email2dict(msgs[1])


def test_babyl_send_no_context(
    monkeypatch: pytest.MonkeyPatch, test_email1: EmailMessage, tmp_path: Path
) -> None:
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "babyl",
            "path": "inbox",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    sender.send(test_email1)
    inbox = Babyl("inbox")
    inbox.lock()
    msgs = list(inbox)
    inbox.close()
    assert len(msgs) == 1
    assert email2dict(test_email1) == email2dict(msgs[0])


def test_babyl_close_unopened(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "babyl",
            "path": "inbox",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    assert isinstance(sender, BabylSender)
    with pytest.raises(ValueError) as excinfo:
        sender.close()
    assert str(excinfo.value) == "Mailbox is not open"



from email import message_from_binary_file, policy
from email.message import EmailMessage
import logging
from typing import Any, IO, List, Optional
import click
from click_loglevel import LogLevel
from dotenv import find_dotenv, load_dotenv
from . import (
    DEFAULT_CONFIG_SECTION,
    __version__,
    from_config_file,
    get_default_configpath,
)
from .errors import Error

NO_SECTION = object()


@click.command()
@click.version_option(
    __version__,
    "-V",
    "--version",
    message="%(prog)s %(version)s",
)
@click.option(
    "-c",
    "--config",
    type=click.Path(dir_okay=False),
    default=get_default_configpath(),
    help="Specify the outgoing configuration file to use",
    show_default=True,
)
@click.option(
    "-E",
    "--env",
    type=click.Path(exists=True, dir_okay=False),
    help="Load environment variables from given .env file",
)
@click.option(
    "-l",
    "--log-level",
    type=LogLevel(),
    default=logging.INFO,
    help="Set logging level  [default: INFO]",
)
@click.option(
    "-s",
    "--section",
    help=(
        "Read configuration from the given key/section of the config file"
        f"  [default: {DEFAULT_CONFIG_SECTION}]"
    ),
    metavar="KEY",
    type=click.UNPROCESSED,
)
@click.option(
    "--no-section",
    "section",
    flag_value=NO_SECTION,
    type=click.UNPROCESSED,
    help="Read configuration from the root of the config file",
)
@click.argument("message", type=click.File("rb"), nargs=-1)
@click.pass_context
def main(
    ctx: click.Context,
    message: List[IO[bytes]],
    config: Optional[str],
    section: Any,
    log_level: int,
    env: Optional[str],
) -> None:
    """
    Common interface for different e-mail methods.

    Visit <https://github.com/jwodder/outgoing> for more information.
    """
    if env is None:
        # dotenv's default behavior doesn't play well with
        # click.testing.CliRunner, so we have to force the library to start
        # searching at the current directory.
        env = find_dotenv(usecwd=True)
    load_dotenv(env)
    logging.basicConfig(
        format="%(asctime)s [%(levelname)-8s] %(name)s %(message)s",
        datefmt="%H:%M:%S",
        level=log_level,
    )
    sectname: Optional[str]
    if section is NO_SECTION:
        sectname = None
    elif section is None:
        sectname = DEFAULT_CONFIG_SECTION
    else:
        assert isinstance(section, str)
        sectname = section
    if not message:
        message = [click.get_binary_stream("stdin")]
    try:
        with from_config_file(config, section=sectname, fallback=False) as sender:
            for fp in message:
                with fp:
                    msg = message_from_binary_file(fp, policy=policy.default)
                assert isinstance(msg, EmailMessage)
                sender.send(msg)
    except Error as e:
        ctx.fail(str(e))


if __name__ == "__main__":
    main()  # pragma: no cover



from pathlib import Path
from typing import Any, Dict, Optional
from pydantic import BaseModel, SecretStr, ValidationError
import pytest
from pytest_mock import MockerFixture
from outgoing.config import Password, StandardPassword


class Config01(BaseModel):
    configpath: Path
    host: str
    username: str
    password: StandardPassword


def test_standard_password(mocker: MockerFixture) -> None:
    m = mocker.patch("outgoing.core.resolve_password", return_value="12345")
    cfg = Config01(
        configpath="foo/bar",
        host="example.com",
        username="me",
        password="sentinel",
    )
    assert cfg.configpath == Path("foo/bar")
    assert cfg.host == "example.com"
    assert cfg.username == "me"
    assert cfg.password == SecretStr("12345")
    m.assert_called_once_with(
        "sentinel",
        host="example.com",
        username="me",
        configpath=Path("foo/bar"),
    )


@pytest.mark.parametrize("badpass", [42, ["SECRET"], True])
def test_standard_password_invalid_type(badpass: Any, mocker: MockerFixture) -> None:
    m = mocker.patch("outgoing.core.resolve_password", return_value="12345")
    with pytest.raises(ValidationError) as excinfo:
        Config01(
            configpath="foo/bar",
            host="example.com",
            username="me",
            password=badpass,
        )
    assert (
        "Password must be either a string or an object with exactly one field"
        in str(excinfo.value)
    )
    m.assert_not_called()


def test_standard_password_invalid_env() -> None:
    with pytest.raises(ValidationError) as excinfo:
        Config01(
            configpath="foo/bar",
            host="example.com",
            username="me",
            password={"env": {"key": "SECRET"}},
        )
    assert "'env' password specifier must be a string" in str(excinfo.value)
    assert "foo/bar" not in str(excinfo.value)


def test_standard_password_invalid_host(mocker: MockerFixture) -> None:
    m = mocker.patch("outgoing.core.resolve_password", return_value="12345")
    with pytest.raises(ValidationError) as excinfo:
        Config01(
            configpath="foo/bar",
            host=[42],
            username="me",
            password="sentinel",
        )
    assert "Insufficient data to determine password" in str(excinfo.value)
    m.assert_not_called()


def test_standard_password_invalid_username(mocker: MockerFixture) -> None:
    m = mocker.patch("outgoing.core.resolve_password", return_value="12345")
    with pytest.raises(ValidationError) as excinfo:
        Config01(
            configpath="foo/bar",
            host="example.com",
            username=[42],
            password="sentinel",
        )
    assert "Insufficient data to determine password" in str(excinfo.value)
    m.assert_not_called()


class Password02(Password):
    @classmethod
    def host(cls, _values: Dict[str, Any]) -> str:
        return "api.example.com"

    @classmethod
    def username(cls, _values: Dict[str, Any]) -> str:
        return "mylogin"


class Config02(BaseModel):
    configpath: Path
    host: str
    username: str
    password: Password02


def test_password_constant_fields(mocker: MockerFixture) -> None:
    m = mocker.patch("outgoing.core.resolve_password", return_value="12345")
    cfg = Config02(
        configpath="foo/bar",
        host="example.com",
        username="me",
        password="sentinel",
    )
    assert cfg.configpath == Path("foo/bar")
    assert cfg.host == "example.com"
    assert cfg.username == "me"
    assert cfg.password == SecretStr("12345")
    m.assert_called_once_with(
        "sentinel",
        host="api.example.com",
        username="mylogin",
        configpath=Path("foo/bar"),
    )


class Password03(Password):
    @classmethod
    def host(cls, values: Dict[str, Any]) -> str:
        return f"http:{values['host']}"

    @classmethod
    def username(cls, values: Dict[str, Any]) -> str:
        return f"{values['username']}@{values['host']}"


class Config03(BaseModel):
    configpath: Path
    host: str
    username: str
    password: Password03


def test_password_callable_fields(mocker: MockerFixture) -> None:
    m = mocker.patch("outgoing.core.resolve_password", return_value="12345")
    cfg = Config03(
        configpath="foo/bar",
        host="example.com",
        username="me",
        password="sentinel",
    )
    assert cfg.configpath == Path("foo/bar")
    assert cfg.host == "example.com"
    assert cfg.username == "me"
    assert cfg.password == SecretStr("12345")
    m.assert_called_once_with(
        "sentinel",
        host="http:example.com",
        username="me@example.com",
        configpath=Path("foo/bar"),
    )


class Config04(BaseModel):
    configpath: Path
    host: str
    username: str
    password: Password


def test_password_unset_fields(mocker: MockerFixture) -> None:
    m = mocker.patch("outgoing.core.resolve_password", return_value="12345")
    cfg = Config04(
        configpath="foo/bar",
        host="example.com",
        username="me",
        password="sentinel",
    )
    assert cfg.configpath == Path("foo/bar")
    assert cfg.host == "example.com"
    assert cfg.username == "me"
    assert cfg.password == SecretStr("12345")
    m.assert_called_once_with(
        "sentinel",
        host=None,
        username=None,
        configpath=Path("foo/bar"),
    )


def test_password_bad_host() -> None:
    with pytest.raises(RuntimeError) as excinfo:
        type("PasswordTest", (Password,), {"host": 42})
    assert str(excinfo.value) == "Password.host must be a str, callable, or None"


def test_password_bad_username() -> None:
    with pytest.raises(RuntimeError) as excinfo:
        type("PasswordTest", (Password,), {"username": 42})
    assert str(excinfo.value) == "Password.username must be a str, callable, or None"


class HostErrorPassword(Password):
    @classmethod
    def host(cls, _values: Dict[str, Any]) -> None:
        raise RuntimeError("Invalid host method")


class HostErrorConfig(BaseModel):
    configpath: Path
    password: HostErrorPassword


def test_host_error_password(mocker: MockerFixture) -> None:
    m = mocker.patch("outgoing.core.resolve_password", return_value="12345")
    with pytest.raises(ValidationError) as excinfo:
        HostErrorConfig(configpath="foo/bar", password="sentinel")
    assert "Insufficient data to determine password" in str(excinfo.value)
    m.assert_not_called()


class UsernameErrorPassword(Password):
    @classmethod
    def username(cls, _values: Dict[str, Any]) -> None:
        raise RuntimeError("Invalid username method")


class UsernameErrorConfig(BaseModel):
    configpath: Path
    password: UsernameErrorPassword


def test_username_error_password(mocker: MockerFixture) -> None:
    m = mocker.patch("outgoing.core.resolve_password", return_value="12345")
    with pytest.raises(ValidationError) as excinfo:
        UsernameErrorConfig(configpath="foo/bar", password="sentinel")
    assert "Insufficient data to determine password" in str(excinfo.value)
    m.assert_not_called()


class OptionalPasswordConfig(BaseModel):
    configpath: Path
    host: str
    username: str
    password: Optional[StandardPassword]


def test_none_optional_password(mocker: MockerFixture) -> None:
    m = mocker.patch("outgoing.core.resolve_password", return_value="12345")
    cfg = OptionalPasswordConfig(
        configpath="foo/bar",
        host="example.com",
        username="me",
        password=None,
    )
    assert cfg.dict() == {
        "configpath": Path("foo/bar"),
        "host": "example.com",
        "username": "me",
        "password": None,
    }
    m.assert_not_called()



from email.message import EmailMessage
import logging
from mailbox import MH
from operator import itemgetter
from pathlib import Path
from typing import List, Union
from mailbits import email2dict
import pytest
from outgoing import Sender, from_dict
from outgoing.senders.mailboxes import MHSender


@pytest.mark.parametrize("folder", [None, "work", ["important", "work"]])
def test_mh_construct(
    folder: Union[str, List[str], None], monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "mh",
            "path": "inbox",
            "folder": folder,
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    assert isinstance(sender, Sender)
    assert isinstance(sender, MHSender)
    assert sender.dict() == {
        "configpath": tmp_path / "foo.txt",
        "path": tmp_path / "inbox",
        "folder": folder,
    }
    assert sender._mbox is None


def test_mh_send_no_folder_new_path(
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    tmp_path: Path,
) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "mh",
            "path": "inbox",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender as s:
        assert sender is s
        sender.send(test_email1)
    inbox = MH("inbox")
    assert inbox.list_folders() == []
    msgs = list(inbox)
    assert len(msgs) == 1
    assert email2dict(test_email1) == email2dict(msgs[0])
    assert caplog.record_tuples == [
        (
            "outgoing.senders.mailboxes",
            logging.DEBUG,
            f"Opening MH mailbox at {tmp_path/'inbox'}, root folder",
        ),
        (
            "outgoing.senders.mailboxes",
            logging.INFO,
            f"Adding e-mail {test_email1['Subject']!r} to MH mailbox at"
            f" {tmp_path/'inbox'}, root folder",
        ),
        (
            "outgoing.senders.mailboxes",
            logging.DEBUG,
            f"Closing MH mailbox at {tmp_path/'inbox'}, root folder",
        ),
    ]


def test_mh_send_folder_str_new_path(
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    tmp_path: Path,
) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "mh",
            "path": "inbox",
            "folder": "work",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender:
        sender.send(test_email1)
    inbox = MH("inbox")
    assert inbox.list_folders() == ["work"]
    work = inbox.get_folder("work")
    msgs = list(work)
    assert len(msgs) == 1
    assert email2dict(test_email1) == email2dict(msgs[0])
    assert caplog.record_tuples == [
        (
            "outgoing.senders.mailboxes",
            logging.DEBUG,
            f"Opening MH mailbox at {tmp_path/'inbox'}, folder 'work'",
        ),
        (
            "outgoing.senders.mailboxes",
            logging.INFO,
            f"Adding e-mail {test_email1['Subject']!r} to MH mailbox at"
            f" {tmp_path/'inbox'}, folder 'work'",
        ),
        (
            "outgoing.senders.mailboxes",
            logging.DEBUG,
            f"Closing MH mailbox at {tmp_path/'inbox'}, folder 'work'",
        ),
    ]


def test_mh_send_folder_list_new_path(
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    tmp_path: Path,
) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "mh",
            "path": "inbox",
            "folder": ["important", "work"],
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender:
        sender.send(test_email1)
    inbox = MH("inbox")
    assert inbox.list_folders() == ["important"]
    important = inbox.get_folder("important")
    assert important.list_folders() == ["work"]
    work = important.get_folder("work")
    msgs = list(work)
    assert len(msgs) == 1
    assert email2dict(test_email1) == email2dict(msgs[0])
    assert caplog.record_tuples == [
        (
            "outgoing.senders.mailboxes",
            logging.DEBUG,
            f"Opening MH mailbox at {tmp_path/'inbox'}, folder 'important'/'work'",
        ),
        (
            "outgoing.senders.mailboxes",
            logging.INFO,
            f"Adding e-mail {test_email1['Subject']!r} to MH mailbox at"
            f" {tmp_path/'inbox'}, folder 'important'/'work'",
        ),
        (
            "outgoing.senders.mailboxes",
            logging.DEBUG,
            f"Closing MH mailbox at {tmp_path/'inbox'}, folder 'important'/'work'",
        ),
    ]


def test_mh_send_no_folder_extant_path(
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    test_email2: EmailMessage,
    tmp_path: Path,
) -> None:
    monkeypatch.chdir(tmp_path)
    inbox = MH("inbox")
    inbox.lock()
    inbox.add(test_email1)
    inbox.unlock()
    sender = from_dict(
        {
            "method": "mh",
            "path": "inbox",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender:
        sender.send(test_email2)
    assert inbox.list_folders() == []
    msgs = list(inbox)
    assert len(msgs) == 2
    msgs.sort(key=itemgetter("Subject"))
    assert email2dict(test_email1) == email2dict(msgs[0])
    assert email2dict(test_email2) == email2dict(msgs[1])


def test_mh_send_folder_str_extant_path(
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    test_email2: EmailMessage,
    tmp_path: Path,
) -> None:
    monkeypatch.chdir(tmp_path)
    inbox = MH("inbox")
    inbox.add(test_email1)
    sender = from_dict(
        {
            "method": "mh",
            "path": "inbox",
            "folder": "work",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender:
        sender.send(test_email2)
    assert inbox.list_folders() == ["work"]
    work = inbox.get_folder("work")
    msgs = list(work)
    assert len(msgs) == 1
    assert email2dict(test_email2) == email2dict(msgs[0])


def test_mh_send_extant_folder_str_extant_path(
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    test_email2: EmailMessage,
    tmp_path: Path,
) -> None:
    monkeypatch.chdir(tmp_path)
    inbox = MH("inbox")
    inbox.add_folder("work").add(test_email1)
    sender = from_dict(
        {
            "method": "mh",
            "path": "inbox",
            "folder": "work",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender:
        sender.send(test_email2)
    assert inbox.list_folders() == ["work"]
    work = inbox.get_folder("work")
    msgs = list(work)
    assert len(msgs) == 2
    msgs.sort(key=itemgetter("Subject"))
    assert email2dict(test_email1) == email2dict(msgs[0])
    assert email2dict(test_email2) == email2dict(msgs[1])


def test_mh_send_folder_list_extant_path(
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    test_email2: EmailMessage,
    tmp_path: Path,
) -> None:
    monkeypatch.chdir(tmp_path)
    inbox = MH("inbox")
    inbox.add(test_email1)
    sender = from_dict(
        {
            "method": "mh",
            "path": "inbox",
            "folder": ["important", "work"],
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender:
        sender.send(test_email2)
    assert inbox.list_folders() == ["important"]
    important = inbox.get_folder("important")
    assert important.list_folders() == ["work"]
    work = important.get_folder("work")
    msgs = list(work)
    assert len(msgs) == 1
    assert email2dict(test_email2) == email2dict(msgs[0])


def test_mh_send_partially_extant_folder_list(
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    test_email2: EmailMessage,
    tmp_path: Path,
) -> None:
    monkeypatch.chdir(tmp_path)
    inbox = MH("inbox")
    inbox.add_folder("important").add(test_email1)
    inbox.add_folder("work")
    sender = from_dict(
        {
            "method": "mh",
            "path": "inbox",
            "folder": ["important", "work"],
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender:
        sender.send(test_email2)
    assert sorted(inbox.list_folders()) == ["important", "work"]
    assert list(inbox.get_folder("work")) == []
    important = inbox.get_folder("important")
    assert important.list_folders() == ["work"]
    work = important.get_folder("work")
    msgs = list(work)
    assert len(msgs) == 1
    assert email2dict(test_email2) == email2dict(msgs[0])


def test_mh_send_no_context(
    monkeypatch: pytest.MonkeyPatch, test_email1: EmailMessage, tmp_path: Path
) -> None:
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "mh",
            "path": "inbox",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    sender.send(test_email1)
    inbox = MH("inbox")
    assert inbox.list_folders() == []
    msgs = list(inbox)
    assert len(msgs) == 1
    assert email2dict(test_email1) == email2dict(msgs[0])


def test_mh_close_unopened(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "mh",
            "path": "inbox",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    assert isinstance(sender, MHSender)
    with pytest.raises(ValueError) as excinfo:
        sender.close()
    assert str(excinfo.value) == "Mailbox is not open"



from typing import Optional
from keyring.backend import KeyringBackend


class Keyring(KeyringBackend):
    def get_password(self, service: str, username: str) -> Optional[str]:
        if service == "api.example.com" and username == "luser":
            return "hunter2"
        else:
            return None

    def set_password(self, service: str, username: str, password: str) -> None:
        pass

    def delete_password(self, service: str, username: str) -> None:
        pass



from email.message import EmailMessage
import logging
import subprocess
from typing import List, Optional, Union
from pydantic import Field
from ..config import Path
from ..util import OpenClosable

log = logging.getLogger(__name__)


class CommandSender(OpenClosable):
    configpath: Optional[Path] = None
    command: Union[str, List[str]] = Field(
        default_factory=lambda: ["sendmail", "-i", "-t"]
    )

    def open(self) -> None:
        pass

    def close(self) -> None:
        pass

    def send(self, msg: EmailMessage) -> None:
        log.info(
            "Sending e-mail %r via command %r",
            msg.get("Subject", "<NO SUBJECT>"),
            self.command,
        )
        subprocess.run(
            self.command,
            shell=isinstance(self.command, str),
            input=bytes(msg),
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )



import pathlib
from typing import Optional
from pydantic import BaseModel, ValidationError
import pytest
from outgoing.config import DirectoryPath, FilePath, Path


class Paths(BaseModel):
    path: Optional[Path]
    filepath: Optional[FilePath]
    dirpath: Optional[DirectoryPath]


def test_path_expanduser(
    tmp_home: pathlib.Path,
) -> None:
    (tmp_home / "foo").mkdir()
    (tmp_home / "foo" / "bar.txt").touch()
    obj = Paths(path="~/nowhere", filepath="~/foo/bar.txt", dirpath="~/foo")
    assert obj.path == tmp_home / "nowhere"
    assert obj.filepath == tmp_home / "foo" / "bar.txt"
    assert obj.dirpath == tmp_home / "foo"


def test_path_default_none() -> None:
    obj = Paths()
    assert obj.path is None
    assert obj.filepath is None
    assert obj.dirpath is None


def test_path_explicit_none() -> None:
    obj = Paths(path=None, filepath=None, dirpath=None)
    assert obj.path is None
    assert obj.filepath is None
    assert obj.dirpath is None


def test_filepath_not_exists(tmp_path: pathlib.Path) -> None:
    with pytest.raises(ValidationError):
        Paths(filepath=tmp_path / "nowhere")


def test_dirpath_not_exists(tmp_path: pathlib.Path) -> None:
    with pytest.raises(ValidationError):
        Paths(dirpath=tmp_path / "nowhere")


def test_filepath_not_file(tmp_path: pathlib.Path) -> None:
    (tmp_path / "foo").mkdir()
    with pytest.raises(ValidationError):
        Paths(filepath=tmp_path / "foo")


def test_dirpath_not_directory(tmp_path: pathlib.Path) -> None:
    (tmp_path / "foo").touch()
    with pytest.raises(ValidationError):
        Paths(dirpath=tmp_path / "foo")


def test_paths_schema_format() -> None:
    schema = Paths.schema()
    assert schema["properties"]["path"]["format"] == "path"
    assert schema["properties"]["filepath"]["format"] == "file-path"
    assert schema["properties"]["dirpath"]["format"] == "directory-path"


class ResolvingPaths(BaseModel):
    configpath: Optional[Path]
    path: Path
    filepath: FilePath
    dirpath: DirectoryPath


def test_path_resolve_to_configpath(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: pathlib.Path,
) -> None:
    (tmp_path / "bar" / "foo").mkdir(parents=True)
    (tmp_path / "bar" / "foo" / "bar.txt").touch()
    monkeypatch.chdir(tmp_path)
    obj = ResolvingPaths(
        configpath="bar/quux.txt",
        path="nowhere",
        filepath="foo/bar.txt",
        dirpath="foo",
    )
    assert obj.configpath == tmp_path / "bar" / "quux.txt"
    assert obj.path == tmp_path / "bar" / "nowhere"
    assert obj.filepath == tmp_path / "bar" / "foo" / "bar.txt"
    assert obj.dirpath == tmp_path / "bar" / "foo"


def test_path_resolve_to_curdir(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: pathlib.Path,
) -> None:
    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "bar.txt").touch()
    monkeypatch.chdir(tmp_path)
    obj = ResolvingPaths(
        configpath=None,
        path="nowhere",
        filepath="foo/bar.txt",
        dirpath="foo",
    )
    assert obj.configpath is None
    assert obj.path == tmp_path / "nowhere"
    assert obj.filepath == tmp_path / "foo" / "bar.txt"
    assert obj.dirpath == tmp_path / "foo"


def test_path_resolve_absolute_configpath(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: pathlib.Path,
) -> None:
    (tmp_path / "foo").mkdir(parents=True)
    (tmp_path / "foo" / "bar.txt").touch()
    monkeypatch.chdir(tmp_path)
    obj = ResolvingPaths(
        configpath="bar/quux.txt",
        path=tmp_path / "nowhere",
        filepath=tmp_path / "foo" / "bar.txt",
        dirpath=tmp_path / "foo",
    )
    assert obj.configpath == tmp_path / "bar" / "quux.txt"
    assert obj.path == tmp_path / "nowhere"
    assert obj.filepath == tmp_path / "foo" / "bar.txt"
    assert obj.dirpath == tmp_path / "foo"



from pathlib import Path
import pytest
from outgoing import resolve_password
from outgoing.errors import InvalidPasswordError


def test_dotenv_password(tmp_path: Path) -> None:
    (tmp_path / "foo.txt").write_text("SECRET=hunter2\n")
    assert (
        resolve_password(
            {"dotenv": {"key": "SECRET", "file": str(tmp_path / "foo.txt")}}
        )
        == "hunter2"
    )


def test_dotenv_password_relative(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    (tmp_path / "foo.txt").write_text("SECRET=hunter2\n")
    monkeypatch.chdir(tmp_path)
    assert (
        resolve_password({"dotenv": {"key": "SECRET", "file": "foo.txt"}}) == "hunter2"
    )


def test_dotenv_password_configpath_relative(tmp_path: Path) -> None:
    (tmp_path / "bar").mkdir()
    (tmp_path / "bar" / "foo.txt").write_text("SECRET=hunter2\n")
    assert (
        resolve_password(
            {"dotenv": {"key": "SECRET", "file": "foo.txt"}},
            configpath=tmp_path / "bar" / "quux.txt",
        )
        == "hunter2"
    )


def test_dotenv_password_expanduser(
    tmp_home: Path,
) -> None:
    (tmp_home / "foo.txt").write_text("SECRET=hunter2\n")
    assert (
        resolve_password({"dotenv": {"key": "SECRET", "file": "~/foo.txt"}})
        == "hunter2"
    )


def test_dotenv_password_expanduser_configpath(
    tmp_home: Path,
) -> None:
    (tmp_home / "foo.txt").write_text("SECRET=hunter2\n")
    (tmp_home / "bar").mkdir()
    (tmp_home / "bar" / "foo.txt").write_text("SECRET=prey3\n")
    assert (
        resolve_password(
            {"dotenv": {"key": "SECRET", "file": "~/foo.txt"}},
            configpath=tmp_home / "bar" / "quux.txt",
        )
        == "hunter2"
    )


def test_dotenv_password_default_configpath(tmp_path: Path) -> None:
    (tmp_path / "bar").mkdir()
    (tmp_path / "bar" / ".env").write_text("SECRET=hunter2\n")
    assert (
        resolve_password(
            {"dotenv": {"key": "SECRET"}}, configpath=tmp_path / "bar" / "quux.txt"
        )
        == "hunter2"
    )


def test_dotenv_password_default_no_configpath(tmp_path: Path) -> None:
    (tmp_path / "bar").mkdir()
    (tmp_path / "bar" / ".env").write_text("SECRET=hunter2\n")
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"dotenv": {"key": "SECRET"}})
    assert (
        str(excinfo.value)
        == "Invalid password configuration: no 'file' or configpath given"
    )


def test_dotenv_password_not_in_file(tmp_path: Path) -> None:
    (tmp_path / "foo.txt").write_text("SECRET=hunter2\n")
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password(
            {"dotenv": {"key": "HIDDEN", "file": str(tmp_path / "foo.txt")}}
        )
    assert (
        str(excinfo.value)
        == f"Invalid password configuration: key 'HIDDEN' not in {tmp_path/'foo.txt'}"
    )


def test_dotenv_password_no_value(tmp_path: Path) -> None:
    (tmp_path / "foo.txt").write_text("SECRET\n")
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password(
            {"dotenv": {"key": "SECRET", "file": str(tmp_path / "foo.txt")}}
        )
    assert (
        str(excinfo.value) == "Invalid password configuration: key 'SECRET' in"
        f" {tmp_path/'foo.txt'} does not have a value"
    )


def test_dotenv_invalid_spec_type() -> None:
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"dotenv": "SECRET"})
    assert (
        str(excinfo.value)
        == "Invalid password configuration: 'dotenv' password specifier must"
        " be an object"
    )


def test_dotenv_no_key() -> None:
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"dotenv": {}})
    assert str(excinfo.value) == (
        "Invalid password configuration: Invalid 'dotenv' password specifier: "
        "1 validation error for DotenvSpec\n"
        "key\n"
        "  field required (type=value_error.missing)"
    )


def test_dotenv_nonexistent_file(tmp_path: Path) -> None:
    dotenv_path = tmp_path / "nowhere"
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"dotenv": {"key": "SECRET", "file": dotenv_path}})
    assert str(excinfo.value) == (
        "Invalid password configuration: Invalid 'dotenv' password specifier:"
        " 1 validation error for DotenvSpec\n"
        "file\n"
        f'  file or directory at path "{dotenv_path}" does not exist'
        f" (type=value_error.path.not_exists; path={dotenv_path})"
    )


def test_dotenv_password_configpath_relative_ignore_spec_configpath(
    tmp_path: Path,
) -> None:
    (tmp_path / "bar").mkdir()
    (tmp_path / "bar" / "foo.txt").write_text("SECRET=hunter2\n")
    assert (
        resolve_password(
            {
                "dotenv": {
                    "key": "SECRET",
                    "file": "foo.txt",
                    "configpath": tmp_path / "gnusto" / "cleesh.txt",
                }
            },
            configpath=tmp_path / "bar" / "quux.txt",
        )
        == "hunter2"
    )



from email.message import EmailMessage
import logging
from pathlib import Path
import subprocess
from typing import List, Union
import pytest
from pytest_mock import MockerFixture
from outgoing import Sender, from_dict
from outgoing.senders.command import CommandSender


def test_command_construct_default(tmp_path: Path) -> None:
    sender = from_dict({"method": "command"}, configpath=tmp_path / "foo.toml")
    assert isinstance(sender, Sender)
    assert isinstance(sender, CommandSender)
    assert sender.dict() == {
        "configpath": tmp_path / "foo.toml",
        "command": ["sendmail", "-i", "-t"],
    }


@pytest.mark.parametrize(
    "command", ["~/my/command --option", ["~/my/command", "--option"]]
)
def test_command_construct(command: Union[str, List[str]], tmp_path: Path) -> None:
    sender = from_dict(
        {"method": "command", "command": command}, configpath=tmp_path / "foo.toml"
    )
    assert isinstance(sender, CommandSender)
    assert sender.dict() == {
        "configpath": tmp_path / "foo.toml",
        "command": command,
    }


@pytest.mark.parametrize(
    "command,shell",
    [
        ("~/my/command --option", True),
        (["~/my/command", "--option"], False),
    ],
)
def test_command_send(
    caplog: pytest.LogCaptureFixture,
    command: Union[str, List[str]],
    shell: bool,
    mocker: MockerFixture,
    test_email1: EmailMessage,
    tmp_path: Path,
) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    m = mocker.patch("subprocess.run")
    sender = from_dict(
        {"method": "command", "command": command}, configpath=tmp_path / "foo.toml"
    )
    with sender as s:
        assert sender is s
        sender.send(test_email1)
    m.assert_called_once_with(
        command,
        shell=shell,
        input=bytes(test_email1),
        check=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    assert caplog.record_tuples == [
        (
            "outgoing.senders.command",
            logging.INFO,
            f"Sending e-mail {test_email1['Subject']!r} via command {command!r}",
        )
    ]


@pytest.mark.parametrize(
    "command,shell",
    [
        ("~/my/command --option", True),
        (["~/my/command", "--option"], False),
    ],
)
def test_command_send_no_context(
    command: Union[str, List[str]],
    shell: bool,
    mocker: MockerFixture,
    test_email1: EmailMessage,
    tmp_path: Path,
) -> None:
    m = mocker.patch("subprocess.run")
    sender = from_dict(
        {"method": "command", "command": command}, configpath=tmp_path / "foo.toml"
    )
    sender.send(test_email1)
    m.assert_called_once_with(
        command,
        shell=shell,
        input=bytes(test_email1),
        check=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )



from email.message import EmailMessage
import logging
from typing import Optional
from ..config import Path
from ..util import OpenClosable

log = logging.getLogger(__name__)


class NullSender(OpenClosable):
    configpath: Optional[Path] = None

    def open(self) -> None:
        pass

    def close(self) -> None:
        pass

    def send(self, msg: EmailMessage) -> None:
        log.info("Discarding e-mail %r", msg.get("Subject", "<NO SUBJECT>"))



import errno
import os
from pathlib import Path
import pytest
from outgoing import resolve_password
from outgoing.errors import InvalidPasswordError


def test_file_password(tmp_path: Path) -> None:
    (tmp_path / "foo.txt").write_text(" hunter2\n")
    assert resolve_password({"file": str(tmp_path / "foo.txt")}) == "hunter2"


def test_file_password_relative(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    (tmp_path / "foo.txt").write_text(" hunter2\n")
    monkeypatch.chdir(tmp_path)
    assert resolve_password({"file": "foo.txt"}) == "hunter2"


def test_file_password_configpath_relative(tmp_path: Path) -> None:
    (tmp_path / "bar").mkdir()
    (tmp_path / "bar" / "foo.txt").write_text(" hunter2\n")
    assert (
        resolve_password({"file": "foo.txt"}, configpath=tmp_path / "bar" / "quux.txt")
        == "hunter2"
    )


def test_file_password_expanduser(
    tmp_home: Path,
) -> None:
    (tmp_home / "foo.txt").write_text(" hunter2\n")
    assert resolve_password({"file": "~/foo.txt"}) == "hunter2"


def test_file_password_expanduser_configpath(
    tmp_home: Path,
) -> None:
    (tmp_home / "foo.txt").write_text(" hunter2\n")
    (tmp_home / "bar").mkdir()
    (tmp_home / "bar" / "foo.txt").write_text("prey3\n")
    assert (
        resolve_password(
            {"file": "~/foo.txt"},
            configpath=tmp_home / "bar" / "quux.txt",
        )
        == "hunter2"
    )


def test_file_password_not_string() -> None:
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"file": ["foo.txt"]})
    assert str(excinfo.value) == (
        "Invalid password configuration: 'file' password specifier must be a string"
    )


def test_file_password_not_string_configpath() -> None:
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"file": ["foo.txt"]}, configpath="bar.txt")
    assert str(excinfo.value) == (
        "bar.txt: Invalid password configuration: 'file' password specifier"
        " must be a string"
    )


def test_file_nonexistent_file(tmp_path: Path) -> None:
    file_path = tmp_path / "nowhere"
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"file": str(file_path)})
    e = FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(file_path))
    assert str(excinfo.value) == (
        f"Invalid password configuration: Invalid 'file' path: {e}"
    )



from email.message import EmailMessage
from pathlib import Path
from traceback import format_exception
from click.testing import CliRunner, Result
from mailbits import email2dict
from pytest_mock import MockerFixture
from outgoing.__main__ import main


def show_result(r: Result) -> str:
    if r.exception is not None:
        assert isinstance(r.exc_info, tuple)
        return "".join(format_exception(*r.exc_info))
    else:
        return r.output


def test_main_stdin(mocker: MockerFixture, test_email1: EmailMessage) -> None:
    m = mocker.patch("outgoing.senders.null.NullSender", autospec=True)
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("cfg.toml").write_text('[outgoing]\nmethod = "null"\n')
        r = runner.invoke(
            main,
            ["--config", "cfg.toml"],
            input=bytes(test_email1),
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    m.assert_called_once_with(method="null", configpath=Path("cfg.toml"))
    instance = m.return_value.__enter__.return_value
    assert instance.send.call_count == 1
    sent = instance.send.call_args[0][0]
    assert isinstance(sent, EmailMessage)
    assert email2dict(sent) == email2dict(test_email1)


def test_main_args(
    mocker: MockerFixture, test_email1: EmailMessage, test_email2: EmailMessage
) -> None:
    m = mocker.patch("outgoing.senders.null.NullSender", autospec=True)
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("cfg.toml").write_text('[outgoing]\nmethod = "null"\n')
        Path("msg1.eml").write_bytes(bytes(test_email1))
        Path("msg2.eml").write_bytes(bytes(test_email2))
        r = runner.invoke(
            main,
            ["--config", "cfg.toml", "msg1.eml", "msg2.eml"],
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    m.assert_called_once_with(method="null", configpath=Path("cfg.toml"))
    instance = m.return_value.__enter__.return_value
    assert instance.send.call_count == 2
    sent1 = instance.send.call_args_list[0][0][0]
    assert isinstance(sent1, EmailMessage)
    assert email2dict(sent1) == email2dict(test_email1)
    sent2 = instance.send.call_args_list[1][0][0]
    assert isinstance(sent2, EmailMessage)
    assert email2dict(sent2) == email2dict(test_email2)


def test_main_custom_section(mocker: MockerFixture, test_email1: EmailMessage) -> None:
    m = mocker.patch("outgoing.senders.null.NullSender", autospec=True)
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("cfg.toml").write_text(
            "[outgoing]\n" 'method = "command"\n' "\n" "[test]\n" 'method = "null"\n'
        )
        r = runner.invoke(
            main,
            ["--config", "cfg.toml", "--section", "test"],
            input=bytes(test_email1),
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    m.assert_called_once_with(method="null", configpath=Path("cfg.toml"))
    instance = m.return_value.__enter__.return_value
    assert instance.send.call_count == 1
    sent = instance.send.call_args[0][0]
    assert isinstance(sent, EmailMessage)
    assert email2dict(sent) == email2dict(test_email1)


def test_main_no_section(mocker: MockerFixture, test_email1: EmailMessage) -> None:
    m = mocker.patch("outgoing.senders.null.NullSender", autospec=True)
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("cfg.toml").write_text(
            'method = "null"\n\n[outgoing]\nmethod = "command"\n'
        )
        r = runner.invoke(
            main,
            ["--config", "cfg.toml", "--no-section"],
            input=bytes(test_email1),
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    m.assert_called_once_with(
        method="null", configpath=Path("cfg.toml"), outgoing={"method": "command"}
    )
    instance = m.return_value.__enter__.return_value
    assert instance.send.call_count == 1
    sent = instance.send.call_args[0][0]
    assert isinstance(sent, EmailMessage)
    assert email2dict(sent) == email2dict(test_email1)


def test_main_default_dotenv(mocker: MockerFixture, test_email1: EmailMessage) -> None:
    m = mocker.patch("smtplib.SMTP", autospec=True)
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path(".env").write_text("DOTENV_PASSWORD=hunter2\n")
        Path("cfg.toml").write_text(
            "[outgoing]\n"
            'method = "smtp"\n'
            'host = "mx.example.com"\n'
            'username = "luser"\n'
            'password = { env = "DOTENV_PASSWORD" }\n'
        )
        r = runner.invoke(
            main,
            ["--config", "cfg.toml"],
            input=bytes(test_email1),
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    assert m.call_args_list == [mocker.call("mx.example.com", 25)]
    assert m.return_value.method_calls == [
        mocker.call.login("luser", "hunter2"),
        mocker.call.send_message(mocker.ANY),
        mocker.call.quit(),
    ]


def test_main_custom_dotenv(mocker: MockerFixture, test_email1: EmailMessage) -> None:
    m = mocker.patch("smtplib.SMTP", autospec=True)
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path(".env").write_text("CUSTOM_DOTENV_PASSWORD=12345\n")
        Path("custom.env").write_text("CUSTOM_DOTENV_PASSWORD=hunter2\n")
        Path("cfg.toml").write_text(
            "[outgoing]\n"
            'method = "smtp"\n'
            'host = "mx.example.com"\n'
            'username = "luser"\n'
            'password = { env = "CUSTOM_DOTENV_PASSWORD" }\n'
        )
        r = runner.invoke(
            main,
            ["--config", "cfg.toml", "--env", "custom.env"],
            input=bytes(test_email1),
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    assert m.call_args_list == [mocker.call("mx.example.com", 25)]
    assert m.return_value.method_calls == [
        mocker.call.login("luser", "hunter2"),
        mocker.call.send_message(mocker.ANY),
        mocker.call.quit(),
    ]



import pytest
from outgoing import resolve_password
from outgoing.errors import InvalidPasswordError


def test_base64_password() -> None:
    assert resolve_password({"base64": "xaHDqcOn4bmbxJPFpw=="}) == "šéçṛēŧ"


def test_base64_password_not_string() -> None:
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"base64": 42})
    assert str(excinfo.value) == (
        "Invalid password configuration: 'base64' password specifier must be a string"
    )


@pytest.mark.parametrize(
    "badpass",
    [
        "not&base64",
        "xaHDqcOn4bmbxJPFpw",  # missing characters
        "xaHDqcOn4bmbxJPFpw===",  # extra characters
        "/u36zg==",  # not UTF-8
    ],
)
def test_base64_password_invalid(badpass: str) -> None:
    with pytest.raises(
        InvalidPasswordError,
        match="^Invalid password configuration: Could not decode base64 password: ",
    ):
        resolve_password({"base64": badpass})



import json
from pathlib import Path
import pytest
from outgoing import from_config_file, get_default_configpath
from outgoing.errors import InvalidConfigError, MissingConfigError
from outgoing.senders.mailboxes import MaildirSender, MboxSender


def test_from_default_config_file(tmp_home: Path) -> None:
    defconf = get_default_configpath()
    defconf.relative_to(tmp_home)  # Lack of error is pre-3.9's is_relative_to
    defconf.parent.mkdir(parents=True, exist_ok=True)
    defconf.write_text('[outgoing]\nmethod = "mbox"\npath = "inbox"\n')
    sender = from_config_file()
    assert isinstance(sender, MboxSender)
    assert sender.dict() == {
        "configpath": defconf,
        "path": defconf.with_name("inbox"),
    }


def test_from_custom_config_file(tmp_home: Path) -> None:
    defconf = get_default_configpath()
    defconf.parent.mkdir(parents=True, exist_ok=True)
    defconf.write_text('[outgoing]\nmethod = "mbox"\npath = "inbox"\n')
    myconf = tmp_home / "foo.toml"
    myconf.write_text('[outgoing]\nmethod = "maildir"\npath = "dirmail"\n')
    sender = from_config_file(myconf)
    assert isinstance(sender, MaildirSender)
    assert sender.dict() == {
        "configpath": myconf,
        "path": tmp_home / "dirmail",
        "folder": None,
    }


def test_from_nonexistent_custom_config_file(tmp_home: Path) -> None:
    defconf = get_default_configpath()
    defconf.parent.mkdir(parents=True, exist_ok=True)
    defconf.write_text('[outgoing]\nmethod = "mbox"\npath = "inbox"\n')
    sender = from_config_file(tmp_home / "foo.toml")
    assert isinstance(sender, MboxSender)
    assert sender.dict() == {
        "configpath": defconf,
        "path": defconf.with_name("inbox"),
    }


def test_from_nonexistent_custom_config_file_no_fallback(tmp_home: Path) -> None:
    defconf = get_default_configpath()
    defconf.parent.mkdir(parents=True, exist_ok=True)
    defconf.write_text('[outgoing]\nmethod = "mbox"\npath = "inbox"\n')
    with pytest.raises(MissingConfigError) as excinfo:
        from_config_file(tmp_home / "foo.toml", fallback=False)
    assert (
        str(excinfo.value)
        == f"outgoing configuration not found in files: {tmp_home/'foo.toml'}"
    )


def test_from_nonexistent_custom_config_file_no_default(tmp_home: Path) -> None:
    defconf = get_default_configpath()
    with pytest.raises(MissingConfigError) as excinfo:
        from_config_file(tmp_home / "foo.toml")
    assert (
        str(excinfo.value) == f"outgoing configuration not found in files: {defconf},"
        f" {tmp_home/'foo.toml'}"
    )


@pytest.mark.usefixtures("tmp_home")
@pytest.mark.parametrize("fallback", [False, True])
def test_from_nonexistent_default_config_file(fallback: bool) -> None:
    defconf = get_default_configpath()
    with pytest.raises(MissingConfigError) as excinfo:
        from_config_file(fallback=fallback)
    assert str(excinfo.value) == f"outgoing configuration not found in files: {defconf}"


def test_from_custom_json_config_file(tmp_path: Path) -> None:
    myconf = tmp_path / "foo.json"
    with myconf.open("w") as fp:
        json.dump(
            {
                "outgoing": {
                    "method": "maildir",
                    "path": "dirmail",
                },
            },
            fp,
        )
    sender = from_config_file(myconf)
    assert isinstance(sender, MaildirSender)
    assert sender.dict() == {
        "configpath": myconf,
        "path": tmp_path / "dirmail",
        "folder": None,
    }


def test_from_unknown_ext(tmp_path: Path) -> None:
    with pytest.raises(InvalidConfigError) as excinfo:
        from_config_file(tmp_path / "foo.xyz")
    assert (
        str(excinfo.value)
        == f"{tmp_path/'foo.xyz'}: Invalid configuration: Unsupported file extension"
    )


def test_from_custom_section(tmp_path: Path) -> None:
    (tmp_path / "foo.toml").write_text(
        "[outgoing]\n"
        'method = "mbox"\n'
        'path = "inbox"\n'
        "\n"
        "[sender]\n"
        'method = "maildir"\n'
        'path = "dirmail"\n'
    )
    sender = from_config_file(tmp_path / "foo.toml", section="sender")
    assert isinstance(sender, MaildirSender)
    assert sender.dict() == {
        "configpath": tmp_path / "foo.toml",
        "path": tmp_path / "dirmail",
        "folder": None,
    }


def test_from_none_section(tmp_path: Path) -> None:
    (tmp_path / "foo.toml").write_text(
        'method = "maildir"\n'
        'path = "dirmail"\n'
        "\n"
        "[outgoing]\n"
        'method = "mbox"\n'
        'path = "inbox"\n'
    )
    sender = from_config_file(tmp_path / "foo.toml", section=None)
    assert isinstance(sender, MaildirSender)
    assert sender.dict() == {
        "configpath": tmp_path / "foo.toml",
        "path": tmp_path / "dirmail",
        "folder": None,
    }


def test_from_no_section_custom_config_file(tmp_home: Path) -> None:
    defconf = get_default_configpath()
    defconf.parent.mkdir(parents=True, exist_ok=True)
    defconf.write_text('[outgoing]\nmethod = "mbox"\npath = "inbox"\n')
    (tmp_home / "foo.toml").write_text(
        '[sender]\nmethod = "maildir"\npath = "dirmail"\n'
    )
    sender = from_config_file(tmp_home / "foo.toml")
    assert isinstance(sender, MboxSender)
    assert sender.dict() == {
        "configpath": defconf,
        "path": defconf.with_name("inbox"),
    }


def test_from_no_section_custom_config_file_no_fallback(tmp_home: Path) -> None:
    defconf = get_default_configpath()
    defconf.parent.mkdir(parents=True, exist_ok=True)
    defconf.write_text('[outgoing]\nmethod = "mbox"\npath = "inbox"\n')
    (tmp_home / "foo.toml").write_text(
        '[sender]\nmethod = "maildir"\npath = "dirmail"\n'
    )
    with pytest.raises(MissingConfigError) as excinfo:
        from_config_file(tmp_home / "foo.toml", fallback=False)
    assert (
        str(excinfo.value)
        == f"outgoing configuration not found in files: {tmp_home/'foo.toml'}"
    )


def test_from_no_section_in_custom_config_file_or_default(tmp_home: Path) -> None:
    defconf = get_default_configpath()
    defconf.parent.mkdir(parents=True, exist_ok=True)
    defconf.write_text('[sender]\nmethod = "mbox"\npath = "inbox"\n')
    (tmp_home / "foo.toml").write_text(
        '[sender]\nmethod = "maildir"\npath = "dirmail"\n'
    )
    with pytest.raises(MissingConfigError) as excinfo:
        from_config_file(tmp_home / "foo.toml")
    assert (
        str(excinfo.value) == f"outgoing configuration not found in files: {defconf},"
        f" {tmp_home/'foo.toml'}"
    )


@pytest.mark.usefixtures("tmp_home")
@pytest.mark.parametrize("fallback", [False, True])
def test_from_no_section_default_config_file(fallback: bool) -> None:
    defconf = get_default_configpath()
    defconf.parent.mkdir(parents=True, exist_ok=True)
    defconf.write_text('[sender]\nmethod = "mbox"\npath = "inbox"\n')
    with pytest.raises(MissingConfigError) as excinfo:
        from_config_file(fallback=fallback)
    assert str(excinfo.value) == f"outgoing configuration not found in files: {defconf}"


def test_toplevel_not_dict(tmp_path: Path) -> None:
    myconf = tmp_path / "foo.json"
    with myconf.open("w") as fp:
        json.dump(
            [
                {
                    "outgoing": {
                        "method": "maildir",
                        "path": "dirmail",
                    },
                }
            ],
            fp,
        )
    with pytest.raises(InvalidConfigError) as excinfo:
        from_config_file(myconf)
    assert (
        str(excinfo.value)
        == f"{myconf}: Invalid configuration: Top-level structure must be a dict/object"
    )


def test_section_not_dict(tmp_path: Path) -> None:
    (tmp_path / "foo.toml").write_text(
        '[[outgoing]]\nmethod = "mbox"\npath = "inbox"\n'
    )
    with pytest.raises(InvalidConfigError) as excinfo:
        from_config_file(tmp_path / "foo.toml")
    assert (
        str(excinfo.value)
        == f"{tmp_path/'foo.toml'}: Invalid configuration: Section must be a"
        " dict/object"
    )



from email.message import EmailMessage
import logging
import smtplib
import sys
from typing import Any, Dict, Optional
from pydantic import Field, PrivateAttr, validator
from ..config import NetrcConfig
from ..util import OpenClosable

if sys.version_info[:2] >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

STARTTLS = "starttls"

log = logging.getLogger(__name__)


class SMTPSender(NetrcConfig, OpenClosable):
    ssl: Literal[False, True, "starttls"] = False
    port: int = Field(0, ge=0)
    _client: Optional[smtplib.SMTP] = PrivateAttr(None)

    @validator("port", always=True)
    def _set_default_port(
        cls, v: Any, values: Dict[str, Any]  # noqa: B902, U100
    ) -> Any:
        if v == 0:
            ssl = values.get("ssl")
            if ssl is True:
                return 465
            elif ssl == STARTTLS:
                return 587
            else:
                return 25
        else:
            return v

    def open(self) -> None:
        # We need to pass the host & port to the constructor instead of calling
        # connect() later due to <https://bugs.python.org/issue36094>.
        if self.ssl is True:
            log.debug(
                "Connecting to SMTP server at %s, port %d, using TLS",
                self.host,
                self.port,
            )
            self._client = smtplib.SMTP_SSL(self.host, self.port)
        else:
            log.debug("Connecting to SMTP server at %s, port %d", self.host, self.port)
            self._client = smtplib.SMTP(self.host, self.port)
        if self.ssl == STARTTLS:
            log.debug("Enabling STARTTLS")
            self._client.starttls()
        if self.username is not None:
            assert self.password is not None
            log.debug("Logging in as %r", self.username)
            self._client.login(self.username, self.password.get_secret_value())

    def close(self) -> None:
        if self._client is None:
            raise ValueError("SMTPSender is not open")
        log.debug("Closing connection to %s", self.host)
        self._client.quit()
        self._client = None

    def send(self, msg: EmailMessage) -> None:
        with self:
            assert self._client is not None
            log.info("Sending e-mail %r via SMTP", msg.get("Subject", "<NO SUBJECT>"))
            self._client.send_message(msg)



from netrc import NetrcParseError
from pathlib import Path
from typing import Optional
import pytest
from outgoing import lookup_netrc
from outgoing.errors import NetrcLookupError


@pytest.mark.parametrize("username", [None, "myname"])
def test_lookup_netrc(tmp_home: Path, username: Optional[str]) -> None:
    (tmp_home / ".netrc").write_text(
        "machine api.example.com\nlogin myname\npassword hunter2\n"
    )
    (tmp_home / ".netrc").chmod(0o600)
    assert lookup_netrc("api.example.com", username=username) == ("myname", "hunter2")


def test_lookup_netrc_username_mismatch(
    tmp_home: Path,
) -> None:
    (tmp_home / ".netrc").write_text(
        "machine api.example.com\nlogin myname\npassword hunter2\n"
    )
    (tmp_home / ".netrc").chmod(0o600)
    with pytest.raises(NetrcLookupError) as excinfo:
        lookup_netrc("api.example.com", username="myself")
    assert str(excinfo.value) == (
        "Username mismatch in netrc: expected 'myself', but netrc says 'myname'"
    )


@pytest.mark.parametrize("username", [None, "yourname"])
def test_lookup_netrc_path(tmp_home: Path, username: Optional[str]) -> None:
    (tmp_home / ".netrc").write_text(
        "machine api.example.com\nlogin myname\npassword hunter2\n"
    )
    (tmp_home / ".netrc").chmod(0o600)
    (tmp_home / "net.rc").write_text(
        "machine api.example.com\nlogin yourname\npassword 12345\n"
    )
    assert lookup_netrc(
        "api.example.com", username=username, path=tmp_home / "net.rc"
    ) == ("yourname", "12345")


@pytest.mark.parametrize("username", [None, "myname"])
def test_lookup_netrc_no_match(tmp_home: Path, username: Optional[str]) -> None:
    (tmp_home / ".netrc").write_text(
        "machine api.example.com\nlogin myname\npassword hunter2\n"
    )
    (tmp_home / ".netrc").chmod(0o600)
    with pytest.raises(NetrcLookupError) as excinfo:
        lookup_netrc("mx.egg-sample.nil", username=username)
    assert (
        str(excinfo.value)
        == "No entry for 'mx.egg-sample.nil' or default found in netrc file"
    )


@pytest.mark.xfail(
    reason="Tests something that mypy says can happen but doesn't happen in CPython",
    raises=NetrcParseError,
)
@pytest.mark.parametrize("username", [None, "myname"])
def test_lookup_netrc_no_password(tmp_home: Path, username: Optional[str]) -> None:
    (tmp_home / ".netrc").write_text("machine api.example.com\nlogin myname\n")
    (tmp_home / ".netrc").chmod(0o600)
    with pytest.raises(NetrcLookupError) as excinfo:
        lookup_netrc("api.example.com", username=username)
    assert str(excinfo.value) == "No password given in netrc entry"



from abc import ABC, abstractmethod
import os
from pathlib import Path
from types import TracebackType
from typing import Optional, Type, TypeVar, Union
from pydantic import BaseModel, PrivateAttr

AnyPath = Union[str, bytes, "os.PathLike[str]", "os.PathLike[bytes]"]

OC = TypeVar("OC", bound="OpenClosable")


class OpenClosable(ABC, BaseModel):
    """
    An abstract base class for creating reentrant_ context managers.  A
    concrete subclass must define ``open()`` and ``close()`` methods;
    `OpenClosable` will then define ``__enter__`` and ``__exit__`` methods that
    keep track of the depth of nested ``with`` statements, calling ``open()``
    and ``close()`` only when entering & exiting the outermost ``with``.

    .. _reentrant: https://docs.python.org/3/library/contextlib.html
                   #reentrant-cms
    """

    _context_depth: int = PrivateAttr(0)

    @abstractmethod
    def open(self) -> None:
        ...

    @abstractmethod
    def close(self) -> None:
        ...

    def __enter__(self: OC) -> OC:
        if self._context_depth == 0:
            self.open()
        self._context_depth += 1
        return self

    def __exit__(
        self,
        _exc_type: Optional[Type[BaseException]],
        _exc_val: Optional[BaseException],
        _exc_tb: Optional[TracebackType],
    ) -> None:
        self._context_depth -= 1
        if self._context_depth == 0:
            self.close()


def resolve_path(path: AnyPath, basepath: Optional[AnyPath] = None) -> Path:
    """
    Convert a path to a `pathlib.Path` instance and resolve it using the same
    rules for as paths in ``outgoing`` configuration files: expand tildes by
    calling `Path.expanduser()`, prepend the parent directory of ``basepath``
    (usually the value of ``configpath``) to the path if given, and then
    resolve the resulting path to make it absolute.

    :param path path: the path to resolve
    :param path basepath: an optional path to resolve ``path`` relative to
    :rtype: pathlib.Path
    """
    p = Path(os.fsdecode(path)).expanduser()
    if basepath is not None:
        p = Path(os.fsdecode(basepath)).parent / p
    return p.resolve()



from pathlib import Path
from shutil import copyfile
import sys
from keyring.backends.null import Keyring as NullKeyring
import pytest
from pytest_mock import MockerFixture
from outgoing import resolve_password
from outgoing.errors import InvalidPasswordError

HUNTER2KEYRING = Path(__file__).resolve().parent.parent / "data" / "Hunter2Keyring.py"

# To disable reading of keyring's config file:
pytestmark = pytest.mark.usefixtures("tmp_home")


def test_keyring_backend(mocker: MockerFixture) -> None:
    m = mocker.patch.object(NullKeyring, "get_password", return_value="hunter2")
    assert (
        resolve_password(
            {
                "keyring": {
                    "service": "api.example.com",
                    "username": "luser",
                    "backend": "keyring.backends.null.Keyring",
                }
            }
        )
        == "hunter2"
    )
    m.assert_called_once_with("api.example.com", "luser")


def test_keyring_host_username_args(mocker: MockerFixture) -> None:
    m = mocker.patch.object(NullKeyring, "get_password", return_value="hunter2")
    assert (
        resolve_password(
            {"keyring": {"backend": "keyring.backends.null.Keyring"}},
            host="api.example.com",
            username="luser",
        )
        == "hunter2"
    )
    m.assert_called_once_with("api.example.com", "luser")


def test_keyring_host_username_args_overridden(mocker: MockerFixture) -> None:
    m = mocker.patch.object(NullKeyring, "get_password", return_value="hunter2")
    assert (
        resolve_password(
            {
                "keyring": {
                    "service": "api.example.com",
                    "username": "luser",
                    "backend": "keyring.backends.null.Keyring",
                }
            },
            host="mx.example.nil",
            username="myself",
        )
        == "hunter2"
    )
    m.assert_called_once_with("api.example.com", "luser")


def test_keyring_backend_envvar(
    mocker: MockerFixture, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv("PYTHON_KEYRING_BACKEND", "keyring.backends.null.Keyring")
    m = mocker.patch.object(NullKeyring, "get_password", return_value="hunter2")
    assert (
        resolve_password(
            {
                "keyring": {
                    "service": "api.example.com",
                    "username": "luser",
                }
            }
        )
        == "hunter2"
    )
    m.assert_called_once_with("api.example.com", "luser")


def test_keyring_backend_envvar_override(
    mocker: MockerFixture, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv("PYTHON_KEYRING_BACKEND", "keyring.backends.fail.Keyring")
    m = mocker.patch.object(NullKeyring, "get_password", return_value="hunter2")
    assert (
        resolve_password(
            {
                "keyring": {
                    "service": "api.example.com",
                    "username": "luser",
                    "backend": "keyring.backends.null.Keyring",
                }
            }
        )
        == "hunter2"
    )
    m.assert_called_once_with("api.example.com", "luser")


def test_keyring_default_backend(mocker: MockerFixture) -> None:
    keyring = mocker.MagicMock(**{"get_password.return_value": "hunter2"})
    get_keyring = mocker.patch("outgoing.passwords.get_keyring", return_value=keyring)
    assert (
        resolve_password(
            {
                "keyring": {
                    "service": "api.example.com",
                    "username": "luser",
                }
            }
        )
        == "hunter2"
    )
    get_keyring.assert_called_once_with()
    keyring.get_password.assert_called_once_with("api.example.com", "luser")


def test_keyring_no_password(mocker: MockerFixture) -> None:
    m = mocker.patch.object(NullKeyring, "get_password", return_value=None)
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password(
            {
                "keyring": {
                    "service": "api.example.com",
                    "username": "luser",
                    "backend": "keyring.backends.null.Keyring",
                }
            }
        )
    assert (
        str(excinfo.value)
        == "Invalid password configuration: Could not find password for service"
        " 'api.example.com', username 'luser' in keyring"
    )
    m.assert_called_once_with("api.example.com", "luser")


def test_keyring_invalid_spec_type() -> None:
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"keyring": "keyring.backends.null.Keyring"})
    assert (
        str(excinfo.value)
        == "Invalid password configuration: 'keyring' password specifier must"
        " be an object"
    )


def test_keyring_path() -> None:
    sys_path = list(sys.path)
    assert (
        resolve_password(
            {
                "keyring": {
                    "service": "api.example.com",
                    "username": "luser",
                    "backend": "Hunter2Keyring.Keyring",
                    "keyring-path": HUNTER2KEYRING.parent,
                }
            }
        )
        == "hunter2"
    )
    assert sys.path == sys_path


def test_keyring_path_expanduser(tmp_home: Path) -> None:
    (tmp_home / "lib").mkdir()
    copyfile(HUNTER2KEYRING, tmp_home / "lib" / HUNTER2KEYRING.name)
    assert (
        resolve_password(
            {
                "keyring": {
                    "service": "api.example.com",
                    "username": "luser",
                    "backend": "Hunter2Keyring.Keyring",
                    "keyring-path": "~/lib",
                }
            }
        )
        == "hunter2"
    )


def test_keyring_path_relative_to_configpath(tmp_path: Path) -> None:
    (tmp_path / "lib").mkdir()
    copyfile(HUNTER2KEYRING, tmp_path / "lib" / HUNTER2KEYRING.name)
    assert (
        resolve_password(
            {
                "keyring": {
                    "service": "api.example.com",
                    "username": "luser",
                    "backend": "Hunter2Keyring.Keyring",
                    "keyring-path": "../lib",
                }
            },
            configpath=tmp_path / "foo" / "conf.toml",
        )
        == "hunter2"
    )


def test_keyring_no_service() -> None:
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"keyring": {"username": "luser"}})
    assert str(excinfo.value) == (
        "Invalid password configuration: Invalid 'keyring' password specifier: "
        "1 validation error for KeyringSpec\n"
        "service\n"
        "  none is not an allowed value (type=type_error.none.not_allowed)"
    )


def test_keyring_no_username() -> None:
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"keyring": {"service": "api.example.com"}})
    assert str(excinfo.value) == (
        "Invalid password configuration: Invalid 'keyring' password specifier:"
        " 1 validation error for KeyringSpec\n"
        "username\n"
        "  none is not an allowed value (type=type_error.none.not_allowed)"
    )


def test_keyring_nonexistent_keyring_path(tmp_path: Path) -> None:
    keyring_path = tmp_path / "nowhere"
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password(
            {
                "keyring": {
                    "service": "api.example.com",
                    "username": "luser",
                    "keyring-path": keyring_path,
                }
            }
        )
    assert str(excinfo.value) == (
        "Invalid password configuration: Invalid 'keyring' password specifier:"
        " 1 validation error for KeyringSpec\n"
        "keyring-path\n"
        f'  file or directory at path "{keyring_path}" does not exist'
        f" (type=value_error.path.not_exists; path={keyring_path})"
    )


def test_keyring_path_relative_to_configpath_ignore_spec_configpath(
    tmp_path: Path,
) -> None:
    (tmp_path / "lib").mkdir()
    copyfile(HUNTER2KEYRING, tmp_path / "lib" / HUNTER2KEYRING.name)
    assert (
        resolve_password(
            {
                "keyring": {
                    "service": "api.example.com",
                    "username": "luser",
                    "backend": "Hunter2Keyring.Keyring",
                    "keyring-path": "../lib",
                    "configpath": tmp_path / "bar" / "baz.txt",
                }
            },
            configpath=tmp_path / "foo" / "conf.toml",
        )
        == "hunter2"
    )



from pathlib import Path
from typing import Optional, Union
from unittest.mock import sentinel
from pydantic import SecretStr, ValidationError
import pytest
from pytest_mock import MockerFixture
from outgoing.config import NetrcConfig


def test_netrc_config(mocker: MockerFixture) -> None:
    m = mocker.patch(
        "outgoing.core.lookup_netrc",
        return_value=(sentinel.USERNAME, "hunter2"),
    )
    cfg = NetrcConfig(
        netrc=True,
        host="api.example.com",
        username="myname",
    )
    assert cfg.dict() == {
        "configpath": None,
        "netrc": True,
        "host": "api.example.com",
        "username": sentinel.USERNAME,
        "password": SecretStr("hunter2"),
    }
    m.assert_called_once_with("api.example.com", username="myname", path=None)


@pytest.mark.parametrize("username", [None, "myname"])
def test_netrc_config_path(
    mocker: MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    username: Optional[str],
) -> None:
    m = mocker.patch(
        "outgoing.core.lookup_netrc",
        return_value=(sentinel.USERNAME, "hunter2"),
    )
    (tmp_path / "foo.txt").touch()
    monkeypatch.chdir(tmp_path)
    cfg = NetrcConfig(
        netrc="foo.txt",
        host="api.example.com",
        username=username,
    )
    assert cfg.dict() == {
        "configpath": None,
        "netrc": tmp_path / "foo.txt",
        "host": "api.example.com",
        "username": sentinel.USERNAME,
        "password": SecretStr("hunter2"),
    }
    m.assert_called_once_with(
        "api.example.com", username=username, path=tmp_path / "foo.txt"
    )


@pytest.mark.parametrize("username", [None, "myname"])
def test_netrc_config_path_expanduser(
    mocker: MockerFixture, tmp_home: Path, username: Optional[str]
) -> None:
    m = mocker.patch(
        "outgoing.core.lookup_netrc",
        return_value=(sentinel.USERNAME, "hunter2"),
    )
    (tmp_home / "foo.txt").touch()
    cfg = NetrcConfig(
        netrc="~/foo.txt",
        host="api.example.com",
        username=username,
    )
    assert cfg.dict() == {
        "configpath": None,
        "netrc": tmp_home / "foo.txt",
        "host": "api.example.com",
        "username": sentinel.USERNAME,
        "password": SecretStr("hunter2"),
    }
    m.assert_called_once_with(
        "api.example.com", username=username, path=tmp_home / "foo.txt"
    )


@pytest.mark.parametrize("username", [None, "myname"])
def test_netrc_config_path_configpath(
    mocker: MockerFixture, tmp_path: Path, username: Optional[str]
) -> None:
    m = mocker.patch(
        "outgoing.core.lookup_netrc",
        return_value=(sentinel.USERNAME, "hunter2"),
    )
    (tmp_path / "foo.txt").touch()
    cfg = NetrcConfig(
        configpath=tmp_path / "quux.txt",
        netrc="foo.txt",
        host="api.example.com",
        username=username,
    )
    assert cfg.dict() == {
        "configpath": tmp_path / "quux.txt",
        "netrc": tmp_path / "foo.txt",
        "host": "api.example.com",
        "username": sentinel.USERNAME,
        "password": SecretStr("hunter2"),
    }
    m.assert_called_once_with(
        "api.example.com", username=username, path=tmp_path / "foo.txt"
    )


@pytest.mark.parametrize("username", [None, "myname"])
def test_netrc_config_no_such_path(
    mocker: MockerFixture, tmp_path: Path, username: Optional[str]
) -> None:
    m = mocker.patch(
        "outgoing.core.lookup_netrc",
        return_value=(sentinel.USERNAME, "hunter2"),
    )
    with pytest.raises(ValidationError):
        NetrcConfig(
            netrc=tmp_path / "foo.txt",
            host="api.example.com",
            username=username,
        )
    m.assert_not_called()


def test_netrc_config_password_no_username(mocker: MockerFixture) -> None:
    m = mocker.patch(
        "outgoing.core.lookup_netrc",
        return_value=(sentinel.USERNAME, "hunter2"),
    )
    with pytest.raises(ValidationError) as excinfo:
        NetrcConfig(
            netrc=False,
            host="api.example.com",
            username=None,
            password="12345",
        )
    assert "Password cannot be given without username" in str(excinfo.value)
    m.assert_not_called()


@pytest.mark.parametrize("netrc", [True, "foo.txt"])
def test_netrc_config_password_netrc(
    mocker: MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
    netrc: Union[bool, str],
    tmp_path: Path,
) -> None:
    m = mocker.patch(
        "outgoing.core.lookup_netrc",
        return_value=(sentinel.USERNAME, "hunter2"),
    )
    (tmp_path / "foo.txt").touch()
    monkeypatch.chdir(tmp_path)
    with pytest.raises(ValidationError) as excinfo:
        NetrcConfig(
            netrc=netrc,
            host="api.example.com",
            username="myname",
            password="12345",
        )
    assert "netrc cannot be set when a password is present" in str(excinfo.value)
    m.assert_not_called()


def test_netrc_config_username_no_password_no_netrc(mocker: MockerFixture) -> None:
    m = mocker.patch(
        "outgoing.core.lookup_netrc",
        return_value=(sentinel.USERNAME, "hunter2"),
    )
    with pytest.raises(ValidationError) as excinfo:
        NetrcConfig(host="api.example.com", username="myname")
    assert "Username cannot be given without netrc or password" in str(excinfo.value)
    m.assert_not_called()


def test_netrc_config_no_netrc(mocker: MockerFixture) -> None:
    m1 = mocker.patch("outgoing.core.resolve_password", return_value="12345")
    m2 = mocker.patch(
        "outgoing.core.lookup_netrc",
        return_value=(sentinel.USERNAME, "hunter2"),
    )
    cfg = NetrcConfig(
        configpath="foo/bar",
        netrc=False,
        host="api.example.com",
        username="myname",
        password="sentinel",
    )
    assert cfg.username == "myname"
    assert cfg.password == SecretStr("12345")
    m1.assert_called_once_with(
        "sentinel",
        host="api.example.com",
        username="myname",
        configpath=Path("foo/bar").resolve(),
    )
    m2.assert_not_called()


def test_netrc_config_no_netrc_key(mocker: MockerFixture) -> None:
    m1 = mocker.patch("outgoing.core.resolve_password", return_value="12345")
    m2 = mocker.patch(
        "outgoing.core.lookup_netrc",
        return_value=(sentinel.USERNAME, sentinel.PASSWORD),
    )
    cfg = NetrcConfig(
        configpath="foo/bar",
        host="api.example.com",
        username="myname",
        password="sentinel",
    )
    assert cfg.username == "myname"
    assert cfg.password == SecretStr("12345")
    m1.assert_called_once_with(
        "sentinel",
        host="api.example.com",
        username="myname",
        configpath=Path("foo/bar").resolve(),
    )
    m2.assert_not_called()


def test_netrc_config_nothing(mocker: MockerFixture) -> None:
    m = mocker.patch(
        "outgoing.core.lookup_netrc",
        return_value=(sentinel.USERNAME, sentinel.PASSWORD),
    )
    cfg = NetrcConfig(
        configpath="foo/bar",
        netrc=False,
        host="api.example.com",
        username=None,
        password=None,
    )
    assert cfg.username is None
    assert cfg.password is None
    m.assert_not_called()


def test_netrc_config_nothing_no_keys(mocker: MockerFixture) -> None:
    m = mocker.patch(
        "outgoing.core.lookup_netrc",
        return_value=(sentinel.USERNAME, sentinel.PASSWORD),
    )
    cfg = NetrcConfig(
        configpath="foo/bar",
        host="api.example.com",
    )
    assert cfg.username is None
    assert cfg.password is None
    m.assert_not_called()


def test_netrc_config_no_entry(tmp_home: Path) -> None:
    (tmp_home / ".netrc").touch()
    with pytest.raises(ValidationError) as excinfo:
        NetrcConfig(host="api.example.com", netrc=True)
    assert (
        "Error retrieving password from netrc file: No entry for"
        " 'api.example.com' or default found in netrc file" in str(excinfo.value)
    )



from email.message import EmailMessage
import logging
from mailbox import Maildir
from operator import itemgetter
from pathlib import Path
from typing import Optional
from mailbits import email2dict
import pytest
from outgoing import Sender, from_dict
from outgoing.senders.mailboxes import MaildirSender


@pytest.mark.parametrize("folder", [None, "work"])
def test_maildir_construct(
    folder: Optional[str], monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "maildir",
            "path": "inbox",
            "folder": folder,
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    assert isinstance(sender, Sender)
    assert isinstance(sender, MaildirSender)
    assert sender.dict() == {
        "configpath": tmp_path / "foo.txt",
        "path": tmp_path / "inbox",
        "folder": folder,
    }
    assert sender._mbox is None


def test_maildir_send_no_folder_new_path(
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    tmp_path: Path,
) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "maildir",
            "path": "inbox",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender as s:
        assert sender is s
        sender.send(test_email1)
    inbox = Maildir("inbox")
    assert inbox.list_folders() == []
    msgs = list(inbox)
    assert len(msgs) == 1
    assert email2dict(test_email1) == email2dict(msgs[0])
    assert caplog.record_tuples == [
        (
            "outgoing.senders.mailboxes",
            logging.DEBUG,
            f"Opening Maildir at {tmp_path/'inbox'}, root folder",
        ),
        (
            "outgoing.senders.mailboxes",
            logging.INFO,
            f"Adding e-mail {test_email1['Subject']!r} to Maildir at"
            f" {tmp_path/'inbox'}, root folder",
        ),
        (
            "outgoing.senders.mailboxes",
            logging.DEBUG,
            f"Closing Maildir at {tmp_path/'inbox'}, root folder",
        ),
    ]


def test_maildir_send_folder_new_path(
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    tmp_path: Path,
) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "maildir",
            "path": "inbox",
            "folder": "work",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender:
        sender.send(test_email1)
    inbox = Maildir("inbox")
    assert inbox.list_folders() == ["work"]
    work = inbox.get_folder("work")
    msgs = list(work)
    assert len(msgs) == 1
    assert email2dict(test_email1) == email2dict(msgs[0])
    assert caplog.record_tuples == [
        (
            "outgoing.senders.mailboxes",
            logging.DEBUG,
            f"Opening Maildir at {tmp_path/'inbox'}, folder 'work'",
        ),
        (
            "outgoing.senders.mailboxes",
            logging.INFO,
            f"Adding e-mail {test_email1['Subject']!r} to Maildir at"
            f" {tmp_path/'inbox'}, folder 'work'",
        ),
        (
            "outgoing.senders.mailboxes",
            logging.DEBUG,
            f"Closing Maildir at {tmp_path/'inbox'}, folder 'work'",
        ),
    ]


def test_maildir_send_no_folder_extant_path(
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    test_email2: EmailMessage,
    tmp_path: Path,
) -> None:
    monkeypatch.chdir(tmp_path)
    inbox = Maildir("inbox")
    inbox.add(test_email1)
    sender = from_dict(
        {
            "method": "maildir",
            "path": "inbox",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender:
        sender.send(test_email2)
    assert inbox.list_folders() == []
    msgs = list(inbox)
    assert len(msgs) == 2
    msgs.sort(key=itemgetter("Subject"))
    assert email2dict(test_email1) == email2dict(msgs[0])
    assert email2dict(test_email2) == email2dict(msgs[1])


def test_maildir_send_new_folder_extant_path(
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    test_email2: EmailMessage,
    tmp_path: Path,
) -> None:
    monkeypatch.chdir(tmp_path)
    inbox = Maildir("inbox")
    inbox.add(test_email1)
    sender = from_dict(
        {
            "method": "maildir",
            "path": "inbox",
            "folder": "work",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender:
        sender.send(test_email2)
    assert inbox.list_folders() == ["work"]
    work = inbox.get_folder("work")
    msgs = list(work)
    assert len(msgs) == 1
    assert email2dict(test_email2) == email2dict(msgs[0])


def test_maildir_send_extant_folder_extant_path(
    monkeypatch: pytest.MonkeyPatch,
    test_email1: EmailMessage,
    test_email2: EmailMessage,
    tmp_path: Path,
) -> None:
    monkeypatch.chdir(tmp_path)
    inbox = Maildir("inbox")
    inbox.add_folder("work").add(test_email1)
    sender = from_dict(
        {
            "method": "maildir",
            "path": "inbox",
            "folder": "work",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    with sender:
        sender.send(test_email2)
    assert inbox.list_folders() == ["work"]
    work = inbox.get_folder("work")
    msgs = list(work)
    assert len(msgs) == 2
    msgs.sort(key=itemgetter("Subject"))
    assert email2dict(test_email1) == email2dict(msgs[0])
    assert email2dict(test_email2) == email2dict(msgs[1])


def test_maildir_send_no_context(
    monkeypatch: pytest.MonkeyPatch, test_email1: EmailMessage, tmp_path: Path
) -> None:
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "maildir",
            "path": "inbox",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    sender.send(test_email1)
    inbox = Maildir("inbox")
    assert inbox.list_folders() == []
    msgs = list(inbox)
    assert len(msgs) == 1
    assert email2dict(test_email1) == email2dict(msgs[0])


def test_maildir_close_unopened(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.chdir(tmp_path)
    sender = from_dict(
        {
            "method": "maildir",
            "path": "inbox",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    assert isinstance(sender, MaildirSender)
    with pytest.raises(ValueError) as excinfo:
        sender.close()
    assert str(excinfo.value) == "Mailbox is not open"



from typing import List
from pydantic import Field
from outgoing.util import OpenClosable


class OpenCloser(OpenClosable):
    calls: List[str] = Field(default_factory=list)

    def open(self) -> None:
        self.calls.append("open")

    def close(self) -> None:
        self.calls.append("close")


def test_openclosable() -> None:
    oc = OpenCloser()
    assert oc.dict() == {"calls": []}
    assert oc._context_depth == 0
    with oc as oc2:
        assert oc is oc2
        assert oc.calls == ["open"]
        assert oc._context_depth == 1
        with oc:
            assert oc.calls == ["open"]
            assert oc._context_depth == 2
            with oc:
                assert oc.calls == ["open"]
                assert oc._context_depth == 3
            assert oc.calls == ["open"]
            assert oc._context_depth == 2
        assert oc.calls == ["open"]
        assert oc._context_depth == 1
    assert oc.calls == ["open", "close"]
    assert oc._context_depth == 0



from abc import abstractmethod
from email.message import EmailMessage
import logging
import mailbox
from typing import List, Optional, TypeVar, Union
from pydantic import PrivateAttr
from ..config import Path
from ..util import OpenClosable

T = TypeVar("T", bound="MailboxSender")

log = logging.getLogger(__name__)


class MailboxSender(OpenClosable):  # ABC inherited from OpenClosable
    _mbox: Optional[mailbox.Mailbox] = PrivateAttr(None)

    @abstractmethod
    def _makebox(self) -> mailbox.Mailbox:
        ...

    @abstractmethod
    def _describe(self) -> str:
        ...

    def open(self) -> None:
        log.debug("Opening %s", self._describe())
        self._mbox = self._makebox()
        self._mbox.lock()

    def close(self) -> None:
        if self._mbox is None:
            raise ValueError("Mailbox is not open")
        log.debug("Closing %s", self._describe())
        self._mbox.unlock()
        self._mbox.close()
        self._mbox = None

    def send(self, msg: EmailMessage) -> None:
        with self:
            assert self._mbox is not None
            log.info(
                "Adding e-mail %r to %s",
                msg.get("Subject", "<NO SUBJECT>"),
                self._describe(),
            )
            self._mbox.add(msg)


class MboxSender(MailboxSender):
    configpath: Optional[Path] = None
    path: Path

    def _makebox(self) -> mailbox.mbox:
        return mailbox.mbox(self.path)

    def _describe(self) -> str:
        return f"mbox at {self.path}"


class MaildirSender(MailboxSender):
    configpath: Optional[Path] = None
    path: Path
    folder: Optional[str] = None

    def _makebox(self) -> mailbox.Maildir:
        box = mailbox.Maildir(self.path)
        if self.folder is not None:
            try:
                box = box.get_folder(self.folder)
            except mailbox.NoSuchMailboxError:
                box = box.add_folder(self.folder)
        return box

    def _describe(self) -> str:
        if self.folder is None:
            folder = "root folder"
        else:
            folder = f"folder {self.folder!r}"
        return f"Maildir at {self.path}, {folder}"


class MHSender(MailboxSender):
    configpath: Optional[Path] = None
    path: Path
    folder: Union[str, List[str], None] = None

    def _makebox(self) -> mailbox.MH:
        box = mailbox.MH(self.path)
        if self.folder is not None:
            folders: List[str]
            if isinstance(self.folder, str):
                folders = [self.folder]
            else:
                folders = self.folder
            for f in folders:
                try:
                    box = box.get_folder(f)
                except mailbox.NoSuchMailboxError:
                    box = box.add_folder(f)
        return box

    def _describe(self) -> str:
        if self.folder is None:
            folder = "root folder"
        elif isinstance(self.folder, str):
            folder = f"folder {self.folder!r}"
        else:
            folder = "folder " + "/".join(map(repr, self.folder))
        return f"MH mailbox at {self.path}, {folder}"


class MMDFSender(MailboxSender):
    configpath: Optional[Path] = None
    path: Path

    def _makebox(self) -> mailbox.MMDF:
        return mailbox.MMDF(self.path)

    def _describe(self) -> str:
        return f"MMDF mailbox at {self.path}"


class BabylSender(MailboxSender):
    configpath: Optional[Path] = None
    path: Path

    def _makebox(self) -> mailbox.Babyl:
        return mailbox.Babyl(self.path)

    def _describe(self) -> str:
        return f"Babyl mailbox at {self.path}"



from email.message import EmailMessage
import logging
from pathlib import Path
import smtplib
from typing import Union
from mailbits import email2dict
from pydantic import SecretStr
import pytest
from pytest_mock import MockerFixture
from smtpdfix import SMTPDFix
from outgoing import Sender, from_dict
from outgoing.errors import InvalidConfigError
from outgoing.senders.smtp import SMTPSender

smtpdfix_headers = ["x-mailfrom", "x-peer", "x-rcptto"]


def test_smtp_construct_default_ssl(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setenv("MY_PASSWORD", "hunter2")
    sender = from_dict(
        {
            "method": "smtp",
            "host": "mx.example.com",
            "username": "me",
            "password": {"env": "MY_PASSWORD"},
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    assert isinstance(sender, Sender)
    assert isinstance(sender, SMTPSender)
    assert sender.dict() == {
        "configpath": tmp_path / "foo.txt",
        "host": "mx.example.com",
        "username": "me",
        "password": SecretStr("hunter2"),
        "port": 25,
        "ssl": False,
        "netrc": False,
    }
    assert sender._client is None


def test_smtp_construct_no_ssl(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setenv("MY_PASSWORD", "hunter2")
    sender = from_dict(
        {
            "method": "smtp",
            "host": "mx.example.com",
            "username": "me",
            "password": {"env": "MY_PASSWORD"},
            "ssl": False,
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    assert isinstance(sender, SMTPSender)
    assert sender.dict() == {
        "configpath": tmp_path / "foo.txt",
        "host": "mx.example.com",
        "username": "me",
        "password": SecretStr("hunter2"),
        "port": 25,
        "ssl": False,
        "netrc": False,
    }
    assert sender._client is None


def test_smtp_construct_ssl(tmp_path: Path) -> None:
    (tmp_path / "secret").write_text("12345\n")
    sender = from_dict(
        {
            "method": "smtp",
            "host": "mx.example.com",
            "username": "me",
            "password": {"file": "secret"},
            "ssl": True,
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    assert isinstance(sender, SMTPSender)
    assert sender.dict() == {
        "configpath": tmp_path / "foo.txt",
        "host": "mx.example.com",
        "username": "me",
        "password": SecretStr("12345"),
        "port": 465,
        "ssl": True,
        "netrc": False,
    }
    assert sender._client is None


def test_smtp_construct_starttls(tmp_path: Path) -> None:
    (tmp_path / "net.rc").write_text(
        "machine mx.example.com\nlogin me\npassword secret\n"
    )
    sender = from_dict(
        {
            "method": "smtp",
            "host": "mx.example.com",
            "netrc": "net.rc",
            "ssl": "starttls",
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    assert isinstance(sender, SMTPSender)
    assert sender.dict() == {
        "configpath": tmp_path / "foo.txt",
        "host": "mx.example.com",
        "username": "me",
        "password": SecretStr("secret"),
        "port": 587,
        "ssl": "starttls",
        "netrc": tmp_path / "net.rc",
    }
    assert sender._client is None


@pytest.mark.parametrize("ssl", [False, True, "starttls"])
def test_smtp_construct_explicit_port(ssl: Union[bool, str], tmp_path: Path) -> None:
    sender = from_dict(
        {
            "method": "smtp",
            "host": "mx.example.com",
            "port": 1337,
            "ssl": ssl,
        },
        configpath=str(tmp_path / "foo.txt"),
    )
    assert isinstance(sender, SMTPSender)
    assert sender.dict() == {
        "configpath": tmp_path / "foo.txt",
        "host": "mx.example.com",
        "username": None,
        "password": None,
        "port": 1337,
        "ssl": ssl,
        "netrc": False,
    }


def test_smtp_construct_negative_port(tmp_path: Path) -> None:
    with pytest.raises(InvalidConfigError):
        from_dict(
            {
                "method": "smtp",
                "host": "mx.example.com",
                "port": -1,
            },
            configpath=str(tmp_path / "foo.txt"),
        )


def test_smtp_send_no_ssl_no_auth(
    caplog: pytest.LogCaptureFixture, mocker: MockerFixture, test_email1: EmailMessage
) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    m = mocker.patch("smtplib.SMTP", autospec=True)
    sender = from_dict({"method": "smtp", "host": "mx.example.com"})
    with sender:
        sender.send(test_email1)
    assert m.call_args_list == [mocker.call("mx.example.com", 25)]
    assert m.return_value.method_calls == [
        mocker.call.send_message(test_email1),
        mocker.call.quit(),
    ]
    assert caplog.record_tuples == [
        (
            "outgoing.senders.smtp",
            logging.DEBUG,
            "Connecting to SMTP server at mx.example.com, port 25",
        ),
        (
            "outgoing.senders.smtp",
            logging.INFO,
            f"Sending e-mail {test_email1['Subject']!r} via SMTP",
        ),
        (
            "outgoing.senders.smtp",
            logging.DEBUG,
            "Closing connection to mx.example.com",
        ),
    ]


def test_smtp_send_no_ssl_auth(
    caplog: pytest.LogCaptureFixture, mocker: MockerFixture, test_email1: EmailMessage
) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    m = mocker.patch("smtplib.SMTP", autospec=True)
    sender = from_dict(
        {
            "method": "smtp",
            "host": "mx.example.com",
            "username": "luser",
            "password": "54321",
        }
    )
    with sender:
        sender.send(test_email1)
    assert m.call_args_list == [mocker.call("mx.example.com", 25)]
    assert m.return_value.method_calls == [
        mocker.call.login("luser", "54321"),
        mocker.call.send_message(test_email1),
        mocker.call.quit(),
    ]
    assert caplog.record_tuples == [
        (
            "outgoing.senders.smtp",
            logging.DEBUG,
            "Connecting to SMTP server at mx.example.com, port 25",
        ),
        ("outgoing.senders.smtp", logging.DEBUG, "Logging in as 'luser'"),
        (
            "outgoing.senders.smtp",
            logging.INFO,
            f"Sending e-mail {test_email1['Subject']!r} via SMTP",
        ),
        (
            "outgoing.senders.smtp",
            logging.DEBUG,
            "Closing connection to mx.example.com",
        ),
    ]


def test_smtp_send_ssl_no_auth(
    caplog: pytest.LogCaptureFixture, mocker: MockerFixture, test_email1: EmailMessage
) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    m = mocker.patch("smtplib.SMTP_SSL", autospec=True)
    sender = from_dict({"method": "smtp", "host": "mx.example.com", "ssl": True})
    with sender:
        sender.send(test_email1)
    assert m.call_args_list == [mocker.call("mx.example.com", 465)]
    assert m.return_value.method_calls == [
        mocker.call.send_message(test_email1),
        mocker.call.quit(),
    ]
    assert caplog.record_tuples == [
        (
            "outgoing.senders.smtp",
            logging.DEBUG,
            "Connecting to SMTP server at mx.example.com, port 465, using TLS",
        ),
        (
            "outgoing.senders.smtp",
            logging.INFO,
            f"Sending e-mail {test_email1['Subject']!r} via SMTP",
        ),
        (
            "outgoing.senders.smtp",
            logging.DEBUG,
            "Closing connection to mx.example.com",
        ),
    ]


def test_smtp_send_ssl_auth(
    caplog: pytest.LogCaptureFixture,
    mocker: MockerFixture,
    test_email1: EmailMessage,
) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    m = mocker.patch("smtplib.SMTP_SSL", autospec=True)
    sender = from_dict(
        {
            "method": "smtp",
            "host": "mx.example.com",
            "username": "luser",
            "password": "54321",
            "ssl": True,
        }
    )
    with sender:
        sender.send(test_email1)
    assert m.call_args_list == [mocker.call("mx.example.com", 465)]
    assert m.return_value.method_calls == [
        mocker.call.login("luser", "54321"),
        mocker.call.send_message(test_email1),
        mocker.call.quit(),
    ]
    assert caplog.record_tuples == [
        (
            "outgoing.senders.smtp",
            logging.DEBUG,
            "Connecting to SMTP server at mx.example.com, port 465, using TLS",
        ),
        ("outgoing.senders.smtp", logging.DEBUG, "Logging in as 'luser'"),
        (
            "outgoing.senders.smtp",
            logging.INFO,
            f"Sending e-mail {test_email1['Subject']!r} via SMTP",
        ),
        (
            "outgoing.senders.smtp",
            logging.DEBUG,
            "Closing connection to mx.example.com",
        ),
    ]


def test_smtp_send_starttls_no_auth(
    caplog: pytest.LogCaptureFixture, mocker: MockerFixture, test_email1: EmailMessage
) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    m = mocker.patch("smtplib.SMTP", autospec=True)
    sender = from_dict({"method": "smtp", "host": "mx.example.com", "ssl": "starttls"})
    with sender:
        sender.send(test_email1)
    assert m.call_args_list == [mocker.call("mx.example.com", 587)]
    assert m.return_value.method_calls == [
        mocker.call.starttls(),
        mocker.call.send_message(test_email1),
        mocker.call.quit(),
    ]
    assert caplog.record_tuples == [
        (
            "outgoing.senders.smtp",
            logging.DEBUG,
            "Connecting to SMTP server at mx.example.com, port 587",
        ),
        ("outgoing.senders.smtp", logging.DEBUG, "Enabling STARTTLS"),
        (
            "outgoing.senders.smtp",
            logging.INFO,
            f"Sending e-mail {test_email1['Subject']!r} via SMTP",
        ),
        (
            "outgoing.senders.smtp",
            logging.DEBUG,
            "Closing connection to mx.example.com",
        ),
    ]


def test_smtp_send_starttls_auth(
    caplog: pytest.LogCaptureFixture, mocker: MockerFixture, test_email1: EmailMessage
) -> None:
    caplog.set_level(logging.DEBUG, logger="outgoing")
    m = mocker.patch("smtplib.SMTP", autospec=True)
    sender = from_dict(
        {
            "method": "smtp",
            "host": "mx.example.com",
            "username": "luser",
            "password": "54321",
            "ssl": "starttls",
        }
    )
    with sender:
        sender.send(test_email1)
    assert m.call_args_list == [mocker.call("mx.example.com", 587)]
    assert m.return_value.method_calls == [
        mocker.call.starttls(),
        mocker.call.login("luser", "54321"),
        mocker.call.send_message(test_email1),
        mocker.call.quit(),
    ]
    assert caplog.record_tuples == [
        (
            "outgoing.senders.smtp",
            logging.DEBUG,
            "Connecting to SMTP server at mx.example.com, port 587",
        ),
        ("outgoing.senders.smtp", logging.DEBUG, "Enabling STARTTLS"),
        ("outgoing.senders.smtp", logging.DEBUG, "Logging in as 'luser'"),
        (
            "outgoing.senders.smtp",
            logging.INFO,
            f"Sending e-mail {test_email1['Subject']!r} via SMTP",
        ),
        (
            "outgoing.senders.smtp",
            logging.DEBUG,
            "Closing connection to mx.example.com",
        ),
    ]


# Spied-on smtplib classes don't work with smtpdfix for some reason (or at
# all?), so we need to split the tests apart.


def test_smtp_fix_send_no_ssl_no_auth(
    smtpd: SMTPDFix, test_email1: EmailMessage
) -> None:
    sender = from_dict({"method": "smtp", "host": smtpd.hostname, "port": smtpd.port})
    with sender:
        sender.send(test_email1)
    assert len(smtpd.messages) == 1
    msgdict = email2dict(smtpd.messages[0])
    for h in smtpdfix_headers:
        msgdict["headers"].pop(h, None)
    assert email2dict(test_email1) == msgdict


def test_smtp_fix_send_no_ssl_auth(smtpd: SMTPDFix, test_email1: EmailMessage) -> None:
    smtpd.config.enforce_auth = True
    smtpd.config.auth_require_tls = False
    smtpd.config.login_username = "luser"
    smtpd.config.login_password = "hunter2"
    sender = from_dict(
        {
            "method": "smtp",
            "host": smtpd.hostname,
            "port": smtpd.port,
            "username": "luser",
            "password": "hunter2",
        }
    )
    with sender:
        sender.send(test_email1)
    assert len(smtpd.messages) == 1
    msgdict = email2dict(smtpd.messages[0])
    for h in smtpdfix_headers:
        msgdict["headers"].pop(h, None)
    assert email2dict(test_email1) == msgdict


def test_smtp_fix_send_ssl_no_auth(smtpd: SMTPDFix, test_email1: EmailMessage) -> None:
    smtpd.config.use_ssl = True
    sender = from_dict(
        {"method": "smtp", "host": smtpd.hostname, "port": smtpd.port, "ssl": True}
    )
    with sender:
        sender.send(test_email1)
    assert len(smtpd.messages) == 1
    msgdict = email2dict(smtpd.messages[0])
    for h in smtpdfix_headers:
        msgdict["headers"].pop(h, None)
    assert email2dict(test_email1) == msgdict


@pytest.mark.skip(reason="Produces PytestUnraisableExceptionWarning on pytest 6.2.3+")
@pytest.mark.xfail(
    raises=smtplib.SMTPNotSupportedError,
    reason="https://github.com/bebleo/smtpdfix/issues/10",
)
def test_smtp_fix_send_ssl_auth(
    smtpd: SMTPDFix,
    test_email1: EmailMessage,
) -> None:
    smtpd.config.enforce_auth = True
    smtpd.config.login_username = "luser"
    smtpd.config.login_password = "hunter2"
    smtpd.config.use_ssl = True
    sender = from_dict(
        {
            "method": "smtp",
            "host": smtpd.hostname,
            "port": smtpd.port,
            "username": "luser",
            "password": "hunter2",
            "ssl": True,
        }
    )
    with sender:
        sender.send(test_email1)
    assert len(smtpd.messages) == 1
    msgdict = email2dict(smtpd.messages[0])
    for h in smtpdfix_headers:
        msgdict["headers"].pop(h, None)
    assert email2dict(test_email1) == msgdict


def test_smtp_fix_send_starttls_no_auth(
    smtpd: SMTPDFix, test_email1: EmailMessage
) -> None:
    smtpd.config.use_starttls = True
    sender = from_dict(
        {
            "method": "smtp",
            "host": smtpd.hostname,
            "port": smtpd.port,
            "ssl": "starttls",
        }
    )
    with sender:
        sender.send(test_email1)
    assert len(smtpd.messages) == 1
    msgdict = email2dict(smtpd.messages[0])
    for h in smtpdfix_headers:
        msgdict["headers"].pop(h, None)
    assert email2dict(test_email1) == msgdict


def test_smtp_fix_send_starttls_auth(
    smtpd: SMTPDFix, test_email1: EmailMessage
) -> None:
    smtpd.config.use_starttls = True
    smtpd.config.enforce_auth = True
    smtpd.config.login_username = "luser"
    smtpd.config.login_password = "hunter2"
    sender = from_dict(
        {
            "method": "smtp",
            "host": smtpd.hostname,
            "port": smtpd.port,
            "username": "luser",
            "password": "hunter2",
            "ssl": "starttls",
        }
    )
    with sender:
        sender.send(test_email1)
    assert len(smtpd.messages) == 1
    msgdict = email2dict(smtpd.messages[0])
    for h in smtpdfix_headers:
        msgdict["headers"].pop(h, None)
    assert email2dict(test_email1) == msgdict


def test_smtp_send_no_context(mocker: MockerFixture, test_email1: EmailMessage) -> None:
    m = mocker.patch("smtplib.SMTP", autospec=True)
    sender = from_dict({"method": "smtp", "host": "mx.example.com"})
    sender.send(test_email1)
    assert m.call_args_list == [mocker.call("mx.example.com", 25)]
    assert m.return_value.method_calls == [
        mocker.call.send_message(test_email1),
        mocker.call.quit(),
    ]


def test_smtp_close_unopened() -> None:
    sender = from_dict({"method": "smtp", "host": "mx.example.com"})
    assert isinstance(sender, SMTPSender)
    with pytest.raises(ValueError) as excinfo:
        sender.close()
    assert str(excinfo.value) == "SMTPSender is not open"



from collections.abc import Mapping
import pathlib
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Union
import pydantic
from . import core
from .errors import InvalidPasswordError
from .util import AnyPath, resolve_path

if TYPE_CHECKING:
    from pydantic.typing import CallableGenerator

    Path = pathlib.Path
    FilePath = pathlib.Path
    DirectoryPath = pathlib.Path

else:

    class Path(pathlib.Path):
        """
        Converts its input to `pathlib.Path` instances, including expanding
        tildes.  If there is a field named ``configpath`` declared before the
        `Path` field and its value is non-`None`, then the value of the `Path`
        field will be resolved relative to the parent directory of the
        ``configpath`` field; otherwise, it will be resolved relative to the
        current directory.
        """

        @classmethod
        def __get_validators__(cls) -> "CallableGenerator":
            yield path_resolve

    class FilePath(pydantic.FilePath):
        """ Like `Path`, but the path must exist and be a file """

        @classmethod
        def __get_validators__(cls) -> "CallableGenerator":
            yield path_resolve
            yield from super().__get_validators__()

    class DirectoryPath(pydantic.DirectoryPath):
        """ Like `Path`, but the path must exist and be a directory """

        @classmethod
        def __get_validators__(cls) -> "CallableGenerator":
            yield path_resolve
            yield from super().__get_validators__()


# We have to implement Password configuration via explicit subclassing as using
# a function instead à la pydantic's conlist() leads to mypy errors; cf.
# <https://github.com/samuelcolvin/pydantic/issues/975>
class Password(pydantic.SecretStr):
    """
    A subclass of `pydantic.SecretStr` that accepts ``outgoing`` password
    specifiers as input and automatically resolves them using
    `resolve_password()`.  Host, username, and ``configpath`` values are passed
    to `resolve_password()` as follows:

    - If `Password` is subclassed and given a ``host`` class variable naming a
      field, and if the subclass is then used in a model where a field with
      that name is declared before the `Password` subclass field, then when the
      model is instantiated, the value of the named field will be passed as the
      ``host`` argument to `resolve_password()`.  (If the named field is not
      present on the model that uses the subclass, the `Password` field will
      fail validation.)

    - Alternatively, `Password` can be subclassed with ``host`` set to a class
      callable (a classmethod or staticmethod), and when that subclass is used
      in a model being instantiated, the callable will be passed a `dict` of
      all validated fields declared before the password field; the return value
      from the callable will then be passed as the ``host`` argument to
      `resolve_password()`.  (If the callable raises an exception, the
      `Password` field will fail validation.)

    - If `Password` is used in a model without being subclassed, or if ``host``
      is not defined in a subclass, then `None` will be passed as the ``host``
      argument to `resolve_password()`.

    - The ``username`` argument to `resolve_password()` can likewise be defined
      by subclassing `Password` and defining ``username`` appropriately.

    - If there is a field named ``configpath`` declared before the `Password`
      field, then the value of ``configpath`` is passed to
      `resolve_password()`.

    For example, if writing a pydantic model for a sender configuration where
    the host-analogue value is passed in a field named ``"service"`` and for
    which the username is always ``"__token__"``, you would subclass `Password`
    like this:

    .. code:: python

        class MyPassword(outgoing.Password):
            host = "service"

            @staticmethod
            def username(values: Dict[str, Any]) -> str:
                return "__token__"

    and then use it in your model like so:

    .. code:: python

        class MySender(pydantic.BaseModel):
            configpath: Optional[outgoing.Path]
            service: str
            password: MyPassword  # Must come after `configpath` and `service`!
            # ... other fields ...

    Then, when ``MySender`` is instantiated, the input to the ``password``
    field would be automatically resolved by doing (effectively):

    .. code:: python

        my_sender.password = pydantic.SecretStr(
            resolve_password(
                my_sender.password,
                host=my_sender.service,
                username="__token__",
                configpath=my_sender.configpath,
            )
        )

    .. note::

        As this is a subclass of `pydantic.SecretStr`, the value of a
        `Password` field is retrieved by calling its ``get_secret_value()``
        method.
    """

    host: ClassVar[Any] = None
    username: ClassVar[Any] = None

    def __init_subclass__(cls) -> None:
        if (
            cls.host is not None
            and not isinstance(cls.host, str)
            and not callable(cls.host)
        ):
            raise RuntimeError("Password.host must be a str, callable, or None")
        if (
            cls.username is not None
            and not isinstance(cls.username, str)
            and not callable(cls.username)
        ):
            raise RuntimeError("Password.username must be a str, callable, or None")

    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        yield cls._resolve
        yield from super().__get_validators__()

    @classmethod
    def _resolve(cls, v: Any, values: Dict[str, Any]) -> str:
        if not isinstance(v, (str, Mapping)):
            raise ValueError(
                "Password must be either a string or an object with exactly one field"
            )
        if isinstance(cls.host, str):
            try:
                host = values[cls.host]
            except KeyError:
                raise ValueError("Insufficient data to determine password")
        elif callable(cls.host):
            try:
                host = cls.host(values)
            except Exception:
                raise ValueError("Insufficient data to determine password")
        else:
            assert cls.host is None
            host = None
        if isinstance(cls.username, str):
            try:
                username = values[cls.username]
            except KeyError:
                raise ValueError("Insufficient data to determine password")
        elif callable(cls.username):
            try:
                username = cls.username(values)
            except Exception:
                raise ValueError("Insufficient data to determine password")
        else:
            assert cls.username is None
            username = None
        try:
            return core.resolve_password(
                v,
                host=host,
                username=username,
                configpath=values.get("configpath"),
            )
        except InvalidPasswordError as e:
            raise ValueError(e.details)


def path_resolve(v: AnyPath, values: Dict[str, Any]) -> pathlib.Path:
    return resolve_path(v, values.get("configpath"))


class StandardPassword(Password):
    """
    A subclass of `Password` in which ``host`` is set to ``"host"`` and
    ``username`` is set to ``"username"``
    """

    host = "host"
    username = "username"


class NetrcConfig(pydantic.BaseModel):
    """
    A pydantic model usable as a base class for any senders that wish to
    support both ``password`` fields and netrc files.  The model accepts the
    fields ``configpath``, ``netrc`` (a boolean or a file path; defaults to
    `False`), ``host`` (required), ``username`` (optional), and ``password``
    (optional).  When the model is instantiated, if ``password`` is `None` but
    ``netrc`` is true or a filepath, the entry for ``host`` is looked up in
    :file:`~/.netrc` or the given file, and the ``username`` and ``password``
    fields are set to the values found.

    The model will raise a validation error if any of the following are true:

    - ``password`` is set but ``netrc`` is true
    - ``password`` is set but ``username`` is not set
    - ``username`` is set but ``password`` is not set and ``netrc`` is false
    - ``netrc`` is true or a filepath, ``username`` is non-`None`, and the
      username in the netrc file differs from ``username``
    - ``netrc`` is true or a filepath and no entry can be found in the netrc
      file
    """

    configpath: Optional[Path]
    netrc: Union[pydantic.StrictBool, FilePath] = False
    host: str
    username: Optional[str]
    password: Optional[StandardPassword]

    @pydantic.root_validator(skip_on_failure=True)
    def _validate(cls, values: Dict[str, Any]) -> Dict[str, Any]:  # noqa: B902, U100
        if values["password"] is not None:
            if values["netrc"]:
                raise ValueError("netrc cannot be set when a password is present")
            elif values["username"] is None:
                raise ValueError("Password cannot be given without username")
        elif values["netrc"]:
            path: Optional[Path]
            if isinstance(values["netrc"], bool):
                path = None
            else:
                path = values["netrc"]
            try:
                username, password = core.lookup_netrc(
                    values["host"], username=values["username"], path=path
                )
            except Exception as e:
                raise ValueError(f"Error retrieving password from netrc file: {e}")
            values["username"] = username
            values["password"] = StandardPassword(password)
        elif values["username"] is not None:
            raise ValueError("Username cannot be given without netrc or password")
        return values



from email.message import EmailMessage
from pathlib import Path
import pytest


@pytest.fixture()
def tmp_home(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> Path:
    # HOME is used by the netrc module in Python 3.6, even on Windows, so
    # always set it.
    monkeypatch.setenv("HOME", str(tmp_path))
    # At this point, why not always set the Windows var?
    monkeypatch.setenv("USERPROFILE", str(tmp_path))
    # Used by appdirs and keyring:
    monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
    monkeypatch.setenv("LOCALAPPDATA", str(tmp_path))
    return tmp_path


@pytest.fixture()
def test_email1() -> EmailMessage:
    msg = EmailMessage()
    msg["Subject"] = "Meet me"
    msg["To"] = "my.beloved@love.love"
    msg["From"] = "me@here.qq"
    msg.set_content(
        "Oh my beloved!\n"
        "\n"
        "Wilt thou dine with me on the morrow?\n"
        "\n"
        "We're having hot pockets.\n"
        "\n"
        "Love, Me\n"
    )
    return msg


@pytest.fixture()
def test_email2() -> EmailMessage:
    msg = EmailMessage()
    msg["Subject"] = "No."
    msg["To"] = "me@here.qq"
    msg["From"] = "my.beloved@love.love"
    msg.set_content("Hot pockets?  Thou disgusteth me.\n\nPineapple pizza or RIOT.\n")
    return msg



from base64 import b64decode
from collections.abc import Mapping
from contextlib import ExitStack
import os
import sys
from typing import Any, Optional
from dotenv import dotenv_values
from keyring import get_keyring
from keyring.backend import KeyringBackend
from keyring.core import load_keyring
from morecontext import additem
from pydantic import BaseModel, Field, ValidationError
from .config import DirectoryPath, FilePath, Path
from .errors import InvalidPasswordError
from .util import AnyPath, resolve_path


def env_scheme(spec: Any) -> str:
    if not isinstance(spec, str):
        raise InvalidPasswordError("'env' password specifier must be a string")
    try:
        return os.environ[spec]
    except KeyError:
        raise InvalidPasswordError(f"Environment variable {spec!r} not set")


def file_scheme(spec: Any, configpath: Optional[AnyPath] = None) -> str:
    try:
        path = os.fsdecode(spec)
    except Exception:
        raise InvalidPasswordError(
            "'file' password specifier must be a string", configpath=configpath
        )
    filepath = resolve_path(path, configpath)
    try:
        return filepath.read_text().strip()
    except OSError as e:
        raise InvalidPasswordError(f"Invalid 'file' path: {e}", configpath=configpath)


def base64_scheme(spec: Any) -> str:
    if not isinstance(spec, str):
        raise InvalidPasswordError("'base64' password specifier must be a string")
    try:
        return b64decode(spec, validate=True).decode()
    except Exception as e:
        raise InvalidPasswordError(f"Could not decode base64 password: {e}")


class DotenvSpec(BaseModel):
    configpath: Optional[Path]
    key: str
    file: Optional[FilePath]


def dotenv_scheme(spec: Any, configpath: Optional[AnyPath] = None) -> str:
    if not isinstance(spec, Mapping):
        raise InvalidPasswordError(
            "'dotenv' password specifier must be an object", configpath=configpath
        )
    try:
        ds = DotenvSpec(**{**spec, "configpath": configpath})
    except ValidationError as e:
        raise InvalidPasswordError(
            f"Invalid 'dotenv' password specifier: {e}", configpath=configpath
        )
    if ds.file is None:
        if configpath is not None:
            ds.file = resolve_path(".env", basepath=configpath)
        else:
            raise InvalidPasswordError("no 'file' or configpath given")
    env = dotenv_values(ds.file)
    try:
        value = env[ds.key]
    except KeyError:
        raise InvalidPasswordError(
            f"key {ds.key!r} not in {ds.file}", configpath=configpath
        )
    else:
        if value is None:
            raise InvalidPasswordError(
                f"key {ds.key!r} in {ds.file} does not have a value",
                configpath=configpath,
            )
        return value


class KeyringSpec(BaseModel):
    configpath: Optional[Path]
    service: str
    username: str
    backend: Optional[str]
    keyring_path: Optional[DirectoryPath] = Field(alias="keyring-path")


def keyring_scheme(
    spec: Any,
    host: Optional[str],
    username: Optional[str],
    configpath: Optional[AnyPath] = None,
) -> str:
    if not isinstance(spec, Mapping):
        raise InvalidPasswordError(
            "'keyring' password specifier must be an object", configpath=configpath
        )
    try:
        ks = KeyringSpec(
            **{"service": host, "username": username, **spec, "configpath": configpath}
        )
    except ValidationError as e:
        raise InvalidPasswordError(
            f"Invalid 'keyring' password specifier: {e}", configpath=configpath
        )
    with ExitStack() as stack:
        keyring: KeyringBackend
        if ks.backend is not None:
            if ks.keyring_path is not None:
                stack.enter_context(
                    additem(sys.path, str(ks.keyring_path), prepend=True)
                )
            keyring = load_keyring(ks.backend)
            keyring.set_properties_from_env()
        else:
            keyring = get_keyring()
        password = keyring.get_password(ks.service, ks.username)
    if password is None:
        raise InvalidPasswordError(
            f"Could not find password for service {ks.service!r}, username"
            f" {ks.username!r} in keyring",
            configpath=configpath,
        )
    else:
        return password



import pytest
from outgoing import resolve_password
from outgoing.errors import InvalidPasswordError


def test_string_password() -> None:
    assert resolve_password("foo") == "foo"


def test_unknown_scheme() -> None:
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"foo": {}})
    assert str(excinfo.value) == (
        "Invalid password configuration: Unsupported password scheme 'foo'"
    )


def test_multiple_keys() -> None:
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"foo": {}, "env": "SECRET"})
    assert str(excinfo.value) == (
        "Invalid password configuration: Password must be either a string or an"
        " object with exactly one field"
    )


def test_invalid_env_configpath() -> None:
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"env": {"key": "SECRET"}}, configpath="foo.toml")
    assert str(excinfo.value) == (
        "foo.toml: Invalid password configuration: 'env' password specifier"
        " must be a string"
    )



import os
from typing import List, Optional, Sequence
from .util import AnyPath


class Error(Exception):
    """ The superclass for all exceptions raised by ``outgoing`` """

    pass


class InvalidConfigError(Error):
    """ Raised on encountering an invalid configuration structure """

    def __init__(self, details: str, configpath: Optional[AnyPath] = None):
        #: A message about the error
        self.details: str = details
        #: The path to the config file containing the invalid configuration
        self.configpath: Optional[AnyPath] = configpath

    def __str__(self) -> str:
        s = ""
        if self.configpath is not None:
            s += f"{os.fsdecode(self.configpath)}: "
        s += f"Invalid configuration: {self.details}"
        return s


class InvalidPasswordError(InvalidConfigError):
    """
    Raised on encountering an invalid password specifier or when no password
    can be determined from a specifier
    """

    def __str__(self) -> str:
        s = ""
        if self.configpath is not None:
            s += f"{os.fsdecode(self.configpath)}: "
        s += f"Invalid password configuration: {self.details}"
        return s


class MissingConfigError(Error):
    """
    Raised when no configuration section can be found in any config files
    """

    def __init__(self, configpaths: Sequence[AnyPath]):
        #: Paths to the configfiles searched for configuration
        self.configpaths: List[AnyPath] = list(configpaths)

    def __str__(self) -> str:
        return "outgoing configuration not found in files: " + ", ".join(
            map(os.fsdecode, self.configpaths)
        )


class NetrcLookupError(Error):
    """
    Raised by `~outgoing.lookup_netrc()` on failure to find a match in a netrc
    file
    """

    pass


class UnsupportedEmailError(Error):
    """
    Raised by sender objects when asked to send an e-mail that uses features or
    constructs that the sending method does not support
    """

    pass



from typing import Any
import pytest
from outgoing import resolve_password
from outgoing.errors import InvalidPasswordError


def test_env_password(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("FOO", "bar")
    assert resolve_password({"env": "FOO"}) == "bar"


def test_env_password_not_set(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("FOO", raising=False)
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"env": "FOO"})
    assert str(excinfo.value) == (
        "Invalid password configuration: Environment variable 'FOO' not set"
    )


@pytest.mark.parametrize("envspec", [["FOO"], {"key": "SECRET"}, 42])
def test_env_password_not_string(envspec: Any) -> None:
    with pytest.raises(InvalidPasswordError) as excinfo:
        resolve_password({"env": envspec})
    assert str(excinfo.value) == (
        "Invalid password configuration: 'env' password specifier must be a string"
    )



from pathlib import Path
import pytest
from outgoing import from_dict
from outgoing.errors import InvalidConfigError
from outgoing.senders.null import NullSender


def test_missing_method() -> None:
    with pytest.raises(InvalidConfigError) as excinfo:
        from_dict({}, configpath="foo.toml")
    assert (
        str(excinfo.value)
        == "foo.toml: Invalid configuration: Required 'method' field not present"
    )


def test_missing_method_no_configpath() -> None:
    with pytest.raises(InvalidConfigError) as excinfo:
        from_dict({})
    assert (
        str(excinfo.value)
        == "Invalid configuration: Required 'method' field not present"
    )


def test_configpath_field(tmp_path: Path) -> None:
    sender = from_dict(
        {"method": "null", "configpath": tmp_path / "bar.toml"},
        configpath=tmp_path / "foo.toml",
    )
    assert isinstance(sender, NullSender)
    assert sender.configpath == tmp_path / "foo.toml"


def test_configpath_field_no_configpath(tmp_path: Path) -> None:
    sender = from_dict({"method": "null", "configpath": tmp_path / "bar.toml"})
    assert isinstance(sender, NullSender)
    assert sender.configpath is None


def test_unknown_method() -> None:
    with pytest.raises(InvalidConfigError) as excinfo:
        from_dict({"method": "foobar"}, configpath="foo.toml")
    assert (
        str(excinfo.value)
        == "foo.toml: Invalid configuration: Unsupported method 'foobar'"
    )


def test_invalid_config() -> None:
    with pytest.raises(
        InvalidConfigError, match=r"^foo\.toml: Invalid configuration: "
    ):
        from_dict({"method": "mbox"}, configpath="foo.toml")
